@@ -79,7 +79,8 @@ interpret_pmstatus (char *str)
 	  title = str;
 	}
 	
-      set_progress (op_general, (int)percentage, 100);
+      set_entertainment_fun (NULL, (int)percentage, 100);
+      set_entertainment_cancel (NULL, NULL);
     }
 }
 
@@ -201,7 +202,7 @@ notice_apt_worker_failure ()
 
   cancel_all_pending_requests ();
 
-  annoy_user_with_log (_("ai_ni_operation_failed"));
+  what_the_fock_p ();
 }
 
 static void
@@ -402,7 +403,7 @@ send_apt_worker_cmd (int cmd)
 				pending[cmd].data,
 				pending[cmd].len))
     {
-      annoy_user_with_log (_("ai_ni_operation_failed"));
+      what_the_fock_p ();
       cancel_request (cmd);
     }
 }
@@ -565,6 +566,13 @@ apt_worker_set_status_callback (apt_worker_callback *callback, void *data)
   pending[APTCMD_STATUS].done_data = data;
 }
 
+void
+apt_worker_noop (apt_worker_callback *callback, void *data)
+{
+  call_apt_worker (APTCMD_NOOP, APTSTATE_DEFAULT, NULL, 0,
+		   callback, data);
+}
+
 void
 apt_worker_get_package_list (int state,
 			     bool only_user,
@@ -585,56 +593,22 @@ apt_worker_get_package_list (int state,
 		   callback, data);
 }
 
-struct awuc_closure {
-  int state;
-  apt_worker_callback *callback;
-  void *data;
-};
-
-
-static void
-apt_worker_update_cache_cont (bool success, void *clos)
-{
-  awuc_closure *c = (awuc_closure *)clos;
-  apt_worker_callback *callback = c->callback;
-  void *data = c->data;
-  int state = c->state;
-  delete c;
-
-  if (success)
-    {
-
-      request.reset ();
-
-      char *http_proxy = get_http_proxy ();
-      request.encode_string (http_proxy);
-      g_free (http_proxy);
-
-      char *https_proxy = get_https_proxy ();
-      request.encode_string (https_proxy);
-      g_free (https_proxy);
-
-      show_progress (_("ai_nw_updating_list"));
-      call_apt_worker (APTCMD_UPDATE_PACKAGE_CACHE, state, 
-		       request.get_buf (), request.get_len (),
-		       callback, data);
-    }
-  else
-    {
-      annoy_user_with_log (_("ai_ni_update_list_not_successful"));
-      callback (APTCMD_UPDATE_PACKAGE_CACHE, NULL, data);
-    }
-}
-
 void
 apt_worker_update_cache (int state, apt_worker_callback *callback, void *data)
 {
-  awuc_closure *c = new awuc_closure;
-  c->callback = callback;
-  c->data = data;
-  c->state = state;
-
-  ensure_network (apt_worker_update_cache_cont, c);
+  request.reset ();
+  
+  char *http_proxy = get_http_proxy ();
+  request.encode_string (http_proxy);
+  g_free (http_proxy);
+  
+  char *https_proxy = get_https_proxy ();
+  request.encode_string (https_proxy);
+  g_free (https_proxy);
+  
+  call_apt_worker (APTCMD_UPDATE_PACKAGE_CACHE, state, 
+		   request.get_buf (), request.get_len (),
+		   callback, data);
 }
 
 void
@@ -698,65 +672,24 @@ apt_worker_install_check (int state, const char *package,
 		   callback, data);
 }
 
-struct awip_closure {
-  char *package;
-  bool updating;
-  int state;
-  apt_worker_callback *callback;
-  void *data;
-};
-
-static void
-apt_worker_install_package_cont (bool success, void *clos)
-{
-  awip_closure *c = (awip_closure *)clos;
-  char *package = c->package;
-  apt_worker_callback *callback = c->callback;
-  void *data = c->data;
-  bool updating = c->updating;
-  int state = c->state;
-  delete c;
-
-  if (success)
-    {
-      request.reset ();
-      request.encode_string (package);
-
-      char *http_proxy = get_http_proxy ();
-      request.encode_string (http_proxy);
-      g_free (http_proxy);
-
-      char *https_proxy = get_https_proxy ();
-      request.encode_string (https_proxy);
-      g_free (https_proxy);
-
-      set_general_progress_title (updating 
-				  ? _("ai_nw_updating")
-				  : _("ai_nw_installing"));
-      reset_progress_was_cancelled ();
-
-      call_apt_worker (APTCMD_INSTALL_PACKAGE, state,
-		       request.get_buf (), request.get_len (),
-		       callback, data);
-    }
-  else
-    callback (APTCMD_INSTALL_PACKAGE, NULL, data);
-
-  g_free (package);
-}
-
 void
 apt_worker_install_package (int state, const char *package, bool updating,
 			    apt_worker_callback *callback, void *data)
 {
-  awip_closure *c = new awip_closure;
-  c->package = g_strdup (package);
-  c->callback = callback;
-  c->data = data;
-  c->updating = updating;
-  c->state = state;
-
-  ensure_network (apt_worker_install_package_cont, c);
+  request.reset ();
+  request.encode_string (package);
+  
+  char *http_proxy = get_http_proxy ();
+  request.encode_string (http_proxy);
+  g_free (http_proxy);
+  
+  char *https_proxy = get_https_proxy ();
+  request.encode_string (https_proxy);
+  g_free (https_proxy);
+  
+  call_apt_worker (APTCMD_INSTALL_PACKAGE, state,
+		   request.get_buf (), request.get_len (),
+		   callback, data);
 }
 
 void
