@@ -58,6 +58,9 @@ void handle_one_apt_worker_response ();
 void apt_worker_set_status_callback (apt_worker_callback *callback,
 				     void *data);
 
+void apt_worker_noop (apt_worker_callback *callback,
+		      void *data);
+
 void apt_worker_get_package_list (int state,
 				  bool only_user,
 				  bool only_installed,
