@@ -90,6 +90,8 @@ struct apt_response_header {
 
 enum apt_proto_result_code {
   rescode_success,              // (success)
+  rescode_partial_success,
+  rescode_cancelled,
   rescode_failure,              // Operation failed
   rescode_download_failed,      // Download failed
   rescode_package_corrupted,    // Package corrupted
@@ -167,7 +169,6 @@ private:
 // - total (int).      Total amount of work to do.
 
 enum apt_proto_operation {
-  op_updating_cache,
   op_downloading,
   op_general
 };
@@ -215,9 +216,12 @@ enum apt_proto_operation {
 //
 // Response contains:
 //
-// - result_code (int)
+// - catalogue_report (xexp).  The current catalogue configuration with 
+//                             error messages attached.
+// - result_code (int).      
 //
-// Error messages appear on stdout/stderr of the apt-worker process.
+// Error messages appear in the catalogue_report as well as on
+// stdout/stderr of the apt-worker process.
 
 
 // GET_SOURCES_LIST - read the main sources.list files, unparsed.
@@ -285,7 +289,8 @@ enum apt_proto_able_status {
   status_corrupted,
   status_incompatible,             // incompatible in general
   status_incompatible_current,     // incompatible with current OS
-  status_system_update_unremovable // could be removed but it's a bad idea
+  status_system_update_unremovable,// could be removed but it's a bad idea
+  status_not_found                 // there is no such package
 };
 
 enum apt_proto_install_flags {
