@@ -25,6 +25,7 @@
 #include <gtk/gtk.h>
 #include <assert.h>
 #include <libintl.h>
+#include <string.h>
 
 #include "details.h"
 #include "util.h"
@@ -34,14 +35,8 @@
 
 #define _(x) gettext (x)
 
-struct spd_closure {
-  package_info *pi;
-  detail_kind kind;
-  bool show_problems;
-  int state;
-  void (*cont) (void * data);
-  void *data;
-};
+/* Utilities
+ */
 
 static const char *
 deptype_name (apt_proto_deptype dep)
@@ -193,7 +188,7 @@ decode_summary (apt_proto_decoder *dec, package_info *pi, detail_kind kind)
 static void
 add_table_field (GtkWidget *table, int row,
 		 const char *field, const char *value,
-		 PangoEllipsizeMode em = PANGO_ELLIPSIZE_END)
+		 PangoEllipsizeMode em = PANGO_ELLIPSIZE_NONE)
 {
   GtkWidget *label;
 
@@ -232,30 +227,137 @@ add_table_list (GtkWidget *table, int row,
   return row;
 }
 
-struct details_closure {
-  void (*cont) (void *data);
+void
+nicify_description_in_place (char *desc)
+{
+  if (desc == NULL)
+    return;
+
+  /* The nicifications are this:
+     
+     - the first space of a line is removed.
+
+     - if after that a line consists solely of a '.', that dot is
+       removed.
+  */
+
+  char *src = desc, *dst = desc;
+  char *bol = src;
+
+  while (*src)
+    {
+      if (src == bol && *src == ' ')
+	{
+	  src++;
+	  continue;
+	}
+
+      if (*src == '\n')
+	{
+	  if (bol+2 == src && bol[0] == ' ' && bol[1] == '.')
+	    dst--;
+	  bol = src + 1;
+	}
+
+      *dst++ = *src++;
+    }
+
+  *dst = '\0';
+}
+
+/* Show package details
+ */
+
+struct spd_clos {
+  package_info *pi;
+  detail_kind kind;
+  bool show_problems;
+  int state;
+  void (*cont) (void * data);
   void *data;
 };
 
-static void
-details_response (GtkDialog *dialog, gint response, gpointer clos)
+static void spd_with_info (package_info *pi, void *data, bool changed);
+static void spd_details_reply (int cmd, apt_proto_decoder *dec, void *data);
+static void spd_with_details (void *data);
+static void spd_response (GtkDialog *dialog, gint response, gpointer data);
+static void spd_end (void *data);
+
+void
+show_package_details (package_info *pi, detail_kind kind,
+		      bool show_problems, int state,
+		      void (*cont) (void *data), void *data)
 {
-  details_closure * closure = (details_closure *) clos;
-  pop_dialog_parent (GTK_WIDGET (dialog));
-  gtk_widget_destroy (GTK_WIDGET (dialog));
-  if (closure->cont)
-    closure->cont(closure->data);
-  delete closure;
+  spd_clos *c = new spd_clos;
+  c->pi = pi;
+  c->kind = kind;
+  c->show_problems = show_problems;
+  c->state = state;
+  c->cont = cont;
+  c->data = data;
+  pi->ref ();
+  get_intermediate_package_info (pi, false, spd_with_info, c, c->state);
+}
+
+void
+spd_with_info (package_info *pi, void *data, bool changed)
+{
+  spd_clos *c = (spd_clos *)data;
+
+  if (pi->have_detail_kind != c->kind)
+    apt_worker_get_package_details (pi->name, (c->kind == remove_details
+					       ? pi->installed_version
+					       : pi->available_version),
+				    c->kind, c->state,
+				    spd_details_reply, c);
+  else
+    spd_with_details (c);
 }
 
 static void
-show_with_details_with_cont (package_info *pi, bool show_problems,
-			     void (*cont) (void *data), void *data)
+spd_details_reply (int cmd, apt_proto_decoder *dec, void *data)
 {
+  spd_clos *c = (spd_clos *)data;
+
+  if (dec == NULL)
+    {
+      spd_end (c);
+      return;
+    }
+
+  g_free (c->pi->maintainer);
+  g_free (c->pi->description);
+  g_free (c->pi->dependencies);
+
+  c->pi->maintainer = dec->decode_string_dup ();
+  c->pi->description = dec->decode_string_dup ();
+  nicify_description_in_place (c->pi->description);
+
+  c->pi->dependencies = decode_dependencies (dec);
+  if (!red_pill_mode || !red_pill_show_deps)
+    {
+      // Too much information can kill you.
+      g_free (c->pi->dependencies);
+      c->pi->dependencies = NULL;
+    }
+      
+  decode_summary (dec, c->pi, c->kind);
+
+  c->pi->have_detail_kind = c->kind;
+
+  spd_with_details (c);
+}
+
+static void
+spd_with_details (void *data)
+{
+  spd_clos *c = (spd_clos *)data;
+
   GtkWidget *dialog, *notebook;
   GtkWidget *table, *common;
   GtkWidget *summary_table, *summary_tab;
-  details_closure *closure = new details_closure;
+
+  package_info *pi = c->pi;
 
   gchar *status;
 
@@ -297,10 +399,20 @@ show_with_details_with_cont (package_info *pi, bool show_problems,
   gtk_table_set_col_spacings (GTK_TABLE (table), 10);
   gtk_table_set_row_spacings (GTK_TABLE (table), 0);
 
-  add_table_field (table, 0,
-		   _("ai_fi_details_package"),
-		   pi->get_display_name (pi->have_detail_kind
-					 == remove_details));
+  {
+    const char *display_name =
+      pi->get_display_name (pi->have_detail_kind == remove_details);
+  
+    if (red_pill_mode && strcmp (pi->name, display_name))
+      {
+	char *extended_name = g_strdup_printf ("%s (%s)", display_name,
+					       pi->name);
+	add_table_field (table, 0, _("ai_fi_details_package"), extended_name);
+	g_free (extended_name);
+      }
+    else
+      add_table_field (table, 0, _("ai_fi_details_package"), display_name);
+  }
 
   gchar *short_description = (pi->have_detail_kind == remove_details
 			      ? pi->installed_short_description
@@ -309,8 +421,7 @@ show_with_details_with_cont (package_info *pi, bool show_problems,
     short_description = pi->installed_short_description;
   add_table_field (table, 1, "", short_description);
 
-  add_table_field (table, 2, _("ai_fi_details_maintainer"),
-		   pi->maintainer, PANGO_ELLIPSIZE_START);
+  add_table_field (table, 2, _("ai_fi_details_maintainer"), pi->maintainer);
 
   add_table_field (table, 3, _("ai_fi_details_status"), status);
 
@@ -343,8 +454,12 @@ show_with_details_with_cont (package_info *pi, bool show_problems,
       add_table_field (table, 8, _("ai_va_details_download_size"), size_buf);
     }
 
-  common = gtk_vbox_new (FALSE, 0);
-  gtk_box_pack_start (GTK_BOX (common), table, TRUE, TRUE, 0);
+  common = gtk_scrolled_window_new (NULL, NULL);
+  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (common),
+					 table);
+  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (common),
+				  GTK_POLICY_AUTOMATIC,
+				  GTK_POLICY_AUTOMATIC);
 
   const gchar *summary_label;
   if (pi->have_detail_kind == remove_details)
@@ -369,7 +484,7 @@ show_with_details_with_cont (package_info *pi, bool show_problems,
   GtkWidget *header_label = make_small_label (pi->summary);
   gtk_misc_set_alignment (GTK_MISC (header_label), 0.0, 0.0);
   gtk_label_set_line_wrap (GTK_LABEL (header_label), FALSE);
-  gtk_label_set_ellipsize (GTK_LABEL (header_label), PANGO_ELLIPSIZE_MIDDLE);
+  gtk_label_set_ellipsize (GTK_LABEL (header_label), PANGO_ELLIPSIZE_NONE);
   gtk_table_attach (GTK_TABLE (summary_table), header_label, 0, 2, 0, 1,
 		    GtkAttachOptions (GTK_EXPAND | GTK_FILL), GTK_FILL,
 		    0, 0);
@@ -414,19 +529,19 @@ show_with_details_with_cont (package_info *pi, bool show_problems,
     }
 
   summary_tab = gtk_scrolled_window_new (NULL, NULL);
-  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW (summary_tab),
-					summary_table);
+  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (summary_tab),
+					 summary_table);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (summary_tab),
 				  GTK_POLICY_AUTOMATIC,
 				  GTK_POLICY_AUTOMATIC);
 
   dialog = gtk_dialog_new_with_buttons (_("ai_ti_details"),
-					get_dialog_parent (),
+					NULL,
 					GTK_DIALOG_MODAL,
 					_("ai_bd_details_close"),
 					GTK_RESPONSE_OK,
 					NULL);
-  push_dialog_parent (dialog);
+  push_dialog (dialog);
 
   gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
   set_dialog_help (dialog, AI_TOPIC ("packagedetailsview"));
@@ -452,132 +567,54 @@ show_with_details_with_cont (package_info *pi, bool show_problems,
 				gtk_label_new ("Dependencies"));
     }
 
-  pi->unref ();
-
-  closure->cont = cont;
-  closure->data = data;
   g_signal_connect (dialog, "response",
-		    G_CALLBACK (details_response), closure);
+		    G_CALLBACK (spd_response), c);
 
   gtk_widget_set_usize (dialog, 600, 320);
   gtk_widget_show_all (dialog);
 
-  if (show_problems)
+  if (c->show_problems)
     gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook),
 				   problems_page);
 }
 
-void
-nicify_description_in_place (char *desc)
+static void
+spd_response (GtkDialog *dialog, gint response, gpointer data)
 {
-  if (desc == NULL)
-    return;
+  spd_clos *c = (spd_clos *)data;
 
-  /* The nicifications are this:
-     
-     - the first space of a line is removed.
+  pop_dialog (GTK_WIDGET (dialog));
+  gtk_widget_destroy (GTK_WIDGET (dialog));
 
-     - if after that a line consists solely of a '.', that dot is
-       removed.
-  */
-
-  char *src = desc, *dst = desc;
-  char *bol = src;
-
-  while (*src)
-    {
-      if (src == bol && *src == ' ')
-	{
-	  src++;
-	  continue;
-	}
-
-      if (*src == '\n')
-	{
-	  if (bol+2 == src && bol[0] == ' ' && bol[1] == '.')
-	    dst--;
-	  bol = src + 1;
-	}
-
-      *dst++ = *src++;
-    }
-
-  *dst = '\0';
+  spd_end (c);
 }
 
 static void
-get_package_details_reply (int cmd, apt_proto_decoder *dec, void *clos)
+spd_end (void *data)
 {
-  spd_closure *c = (spd_closure *)clos;
-  package_info *pi = c->pi;
-  detail_kind kind = c->kind;
-  bool show_problems = c->show_problems;
-  void (*cont) (void *data) = c->cont;
-  void *data = c->data;
+  spd_clos *c = (spd_clos *)data;
+  
+  c->pi->unref ();
+  c->cont (c->data);
+
   delete c;
+}
 
-  if (dec == NULL)
-    {
-      pi->unref ();
-      if (cont != NULL) {
-	cont(data);
-      }
-      return;
-    }
+/* Show package details as an interaction flow
+ */
 
-  g_free (pi->maintainer);
-  g_free (pi->description);
-  g_free (pi->dependencies);
+static void spdf_end (void *data);
 
-  pi->maintainer = dec->decode_string_dup ();
-  pi->description = dec->decode_string_dup ();
-  nicify_description_in_place (pi->description);
-
-  pi->dependencies = decode_dependencies (dec);
-  if (!red_pill_mode || !red_pill_show_deps)
-    {
-      // Too much information can kill you.
-      g_free (pi->dependencies);
-      pi->dependencies = NULL;
-    }
-      
-  decode_summary (dec, pi, kind);
-
-  pi->have_detail_kind = kind;
-
-  show_with_details_with_cont (pi, show_problems, cont, data);
+void
+show_package_details_flow (package_info *pi, detail_kind kind)
+{
+  if (start_interaction_flow ())
+    show_package_details (pi, kind, false, APTSTATE_DEFAULT,
+			  spdf_end, NULL);
 }
 
 void
-spd_cont (package_info *pi, void *data, bool changed)
+spdf_end (void *data)
 {
-  spd_closure *c = (spd_closure *)data;
-
-  if (pi->have_detail_kind != c->kind)
-    apt_worker_get_package_details (pi->name, (c->kind == remove_details
-					       ? pi->installed_version
-					       : pi->available_version),
-				    c->kind, c->state,
-				    get_package_details_reply,
-				    data);
-  else
-    {
-      show_with_details_with_cont (pi, c->show_problems, c->cont, c->data);
-      delete c;
-    }
-}
-
-void
-show_package_details (package_info *pi, detail_kind kind,
-		      bool show_problems, int state)
-{
-  spd_closure *c = new spd_closure;
-  c->pi = pi;
-  c->kind = kind;
-  c->show_problems = show_problems;
-  c->state = state;
-  c->cont = NULL;
-  c->data = NULL;
-  pi->ref ();
-  get_intermediate_package_info (pi, false, spd_cont, c, c->state);
+  end_interaction_flow ();
 }
