@@ -45,6 +45,7 @@
 #include "search.h"
 #include "instr.h"
 #include "repo.h"
+#include "dbus.h"
 
 #define MAX_PACKAGES_NO_CATEGORIES 7
 
@@ -143,6 +144,7 @@ show_view_callback (GtkWidget *btn, gpointer data)
 static gboolean
 view_clicked (HildonBreadCrumbTrail *bct, gpointer node, gpointer user_data)
 {
+  reset_global_target_path ();
   show_view ((view *)((GList*)node)->data);
 
   return TRUE;
@@ -354,6 +356,7 @@ package_info::package_info ()
 {
   ref_count = 1;
   name = NULL;
+  broken = false;
   installed_version = NULL;
   installed_section = NULL;
   installed_pretty_name = NULL;
@@ -550,6 +553,9 @@ canonicalize_section_name (const char *name)
 const char *
 nicify_section_name (const char *name)
 {
+  if (name == NULL)
+    name = "";
+
   if (red_pill_mode && red_pill_show_all)
     return name;
 
@@ -775,7 +781,7 @@ get_package_list_reply_default (int cmd, apt_proto_decoder *dec, void *data)
   if (dec == NULL)
     ;
   else if (dec->decode_int () == 0)
-    annoy_user_with_log (_("ai_ni_operation_failed"));
+    what_the_fock_p ();
   else
     {
       section_info *all_si = new section_info;
@@ -861,7 +867,7 @@ get_package_list_reply_temp (int cmd, apt_proto_decoder *dec, void *data)
   if (dec == NULL)
     ;
   else if (dec->decode_int () == 0)
-    annoy_user_with_log (_("ai_ni_operation_failed"));
+    what_the_fock_p ();
   else
     {
 
@@ -1136,38 +1142,232 @@ get_intermediate_package_info (package_info *pi,
     get_next_package_info (NULL, NULL, false);
 }
 
-struct rpc_clos {
-  bool res;
-  const char *message;
+struct gipl_clos {
+  GList *packages;
+  bool only_installable_info;
   int state;
+  
+  void (*cont) (void *data);
+  void *data;
+};
+
+static void gipl_loop (void *data);
+static void gipl_next (package_info *unused_1, void *data, bool unused_2);
+
+void
+get_intermediate_package_list_info (GList *packages,
+				    bool only_installable_info,
+				    void (*cont) (void *data),
+				    void *data,
+				    int state)
+{
+  gipl_clos *c = new gipl_clos;
+
+  c->packages = packages;
+  c->only_installable_info = only_installable_info;
+  c->state = state;
+  c->cont = cont;
+  c->data = data;
+  
+  gipl_loop (c);
+}
+
+
+static void
+gipl_loop (void *data)
+{
+  gipl_clos *c = (gipl_clos *)data;
+
+  if (c->packages == NULL)
+    {
+      c->cont (c->data);
+      delete c;
+    }
+  else
+    get_intermediate_package_info ((package_info *)(c->packages->data),
+				   c->only_installable_info,
+				   gipl_next, c,
+				   c->state);
+}
+
+static void
+gipl_next (package_info *unused_1, void *data, bool unused_2)
+{
+  gipl_clos *c = (gipl_clos *)data;
+
+  c->packages = c->packages->next;
+  gipl_loop (c);
+}
+
+/** Refreshing the package cache
+
+    0. Optionally set new catalogues
+
+    1. Optionally ask confirmation
+
+    2. Optionally establish network
+
+    3. Perform update_cache operation
+
+    4. Show report with the ability to adjust repositories.  Ask for
+       confirmation and go back to 3 if user changed any repositories.
+
+    5. Reget package list.
+ */
+
+static void rpc_set_catalogues (void *data);
+static void rpc_set_catalogues_reply (int cmd, apt_proto_decoder *dec, 
+				      void *data);
+static void rpc_ask (void *data);
+static void rpc_do_it (bool res, void *data);
+static void rpc_with_network (bool success, void *data);
+static void rpc_update_cache_reply (int cmd, apt_proto_decoder *dec,
+				    void *data);
+static void rpc_show_error_report (void *data);
+static void rpc_error_response (GtkWidget *dialog, gint response,
+				gpointer data);
+static void rpc_detailed_report_done (bool changed, void *data);
+static void rpc_done (void *data);
+
+static void rpc_end (void *data);
+
+static bool refreshed_this_session = false;
+
+struct rpc_clos {
+  int state;
+  bool ask, continued;
+
+  xexp *catalogues;
+  xexp *catalogue_report;
+  apt_proto_result_code result_code;
+  bool keep_going;
+
   void (*cont) (bool res, void *data);
   void *data;
 };
 
-static void
-refresh_package_cache_cont3 (void *data)
+void
+refresh_package_cache (int state,
+		       xexp *new_catalogues,
+		       bool ask, bool continued,
+		       void (*cont) (bool res, void *data), 
+		       void *data)
 {
-  rpc_clos *c = (rpc_clos *) data;
-  if (c->cont)
-    c->cont (c->res, c->data);
-  delete c;
+  refreshed_this_session = true;
+
+  rpc_clos *c = new rpc_clos;
+
+  c->ask = ask;
+  c->continued = continued;
+  c->catalogues = new_catalogues;
+  c->cont = cont;
+  c->data = data;
+  c->state = state;
+
+  c->result_code = rescode_failure;
+  c->keep_going = false;
+
+  rpc_set_catalogues (c);
 }
 
 static void
-refresh_package_cache_cont2 (void *data)
+rpc_set_catalogues (void *data)
 {
-  rpc_clos *c = (rpc_clos *) data;
-  if (!c->res && c->message)
-    annoy_user_with_cont (c->message, refresh_package_cache_cont3, c);
+  rpc_clos *c = (rpc_clos *)data;
+
+  if (c->catalogues)
+    apt_worker_set_catalogues (c->state, c->catalogues,
+			       rpc_set_catalogues_reply, c);
   else
-    refresh_package_cache_cont3 (c);
+    rpc_ask (c);
 }
 
 static void
-refresh_package_cache_reply (int cmd, apt_proto_decoder *dec, void *data)
+rpc_set_catalogues_reply (int cmd, apt_proto_decoder *dec, void *data)
+{
+  rpc_clos *c = (rpc_clos *)data;
+  int success = 0;
+
+  if (dec)
+    success = dec->decode_int ();
+
+  if (!success)
+    {
+      if (dec)
+	what_the_fock_p ();
+      c->result_code = rescode_failure;
+      c->keep_going = false;
+      rpc_done (c);
+    }
+
+  save_backup_data ();
+
+  rpc_ask (c);
+}
+
+static void
+rpc_ask (void *data)
+{
+  rpc_clos *c = (rpc_clos *)data;
+  
+  if (c->ask)
+    ask_yes_no (_("ai_nc_confirm_update"), rpc_do_it, c);
+  else
+    rpc_do_it (true, c);
+}
+
+static void
+rpc_do_it (bool res, void *data)
+{
+  rpc_clos *c = (rpc_clos *)data;
+
+  if (res)
+    {
+      set_entertainment_fun (NULL, -1, 0);
+      set_entertainment_cancel (NULL, NULL);
+      set_entertainment_title (_("ai_nw_updating_list"));
+
+      start_entertaining_user ();
+
+      /* XXX - Only ask for the network when this is for the default
+	       state of the apt-worker.  The temporary state is used
+	       for memory card installs that don't need and don't want
+	       any network.
+
+	       Deciding this here is probably a bit too smart.
+      */
+
+      if (c->state == APTSTATE_DEFAULT)
+	ensure_network (rpc_with_network, c);
+      else
+	rpc_with_network (true, c);
+    }
+  else
+    rpc_done (c);
+}
+
+static void
+rpc_with_network (bool success, void *data)
+{
+  rpc_clos *c = (rpc_clos *)data;
+
+  if (success)
+    apt_worker_update_cache (c->state, rpc_update_cache_reply, c);
+  else
+    {
+      stop_entertaining_user ();
+      
+      c->result_code = rescode_failure;
+      c->keep_going = false;
+      annoy_user (_("No network connection"),
+		  rpc_done, c);
+    }
+}
+
+static void
+rpc_update_cache_reply (int cmd, apt_proto_decoder *dec, void *data)
 {
   rpc_clos *c = (rpc_clos *) data;
-  bool success = true;
 
   /* The updating might have been 'cancelled' or it might have failed,
      and we want to distinguish between those two situations in the
@@ -1181,97 +1381,161 @@ refresh_package_cache_reply (int cmd, apt_proto_decoder *dec, void *data)
      disconnects the network.
   */
 
-  hide_progress ();
+  stop_entertaining_user ();
 
   if (dec == NULL)
     {
       /* Network connection failed or apt-worker crashed.  An error
 	 message has already been displayed.
       */
-      c->message = NULL;
-      success = false;
+      c->result_code = rescode_failure;
+      c->keep_going = false;
+      rpc_done (c);
     }
-  else if (progress_was_cancelled ())
+  else if (entertainment_was_cancelled ())
     {
       /* The user hit cancel.  We don't care whether the operation was
 	 successful or not.
       */
-      c->message = _("ai_ni_update_list_cancelled");
-      success = false;
+      c->result_code = rescode_cancelled;
+      c->keep_going = false;
+      annoy_user (_("ai_ni_update_list_cancelled"), rpc_done, c);
     }
   else
     {
-      int result_code = dec->decode_int ();
+      xexp_free (c->catalogues);
+      c->catalogues = dec->decode_xexp ();
+      c->result_code = apt_proto_result_code (dec->decode_int ());
 
-      if (result_code == rescode_download_failed)
+      if (c->result_code != rescode_success)
+	rpc_show_error_report (c);
+      else
 	{
-	  c->message = _("ai_ni_error_download_failed");
-	  success = false;
-	}
-      else if (result_code != rescode_success)
-	{
-	  c->message = _("ai_ni_update_list_not_successful");
-	  success = false;
+	  c->keep_going = true;
+	  rpc_done (c);
 	}
     }
-
-  c->res = success;
-
-  if (success)
-    {
-      last_update = time (NULL);
-      save_settings ();
-    }
-
-  // We get a new list of available packages even when the refresh
-  // failed because it might have partially succeeded and changed
-  // anyway.  So we need to resynchronize.
-  
-  get_package_list_with_cont (c->state, refresh_package_cache_cont2, c);
 }
 
-static bool refreshed_this_session = false;
-
 static void
-refresh_package_cache_cont (bool res, void *data)
+rpc_show_error_report (void *data)
 {
-  rpc_clos *c = (rpc_clos *)data;
+  rpc_clos *c = (rpc_clos *) data;
 
-  if (res)
+  GtkWidget *dialog;
+  
+  if (c->result_code == rescode_partial_success)
     {
-      apt_worker_update_cache (c->state, refresh_package_cache_reply, c);
+      if (c->continued)
+	dialog = hildon_note_new_confirmation_add_buttons 
+	  (NULL,
+	   _("ai_ni_update_partly_successful"),
+	   _("ai_bd_continue"), GTK_RESPONSE_OK,
+	   _("ai_ni_bd_details"), 1,
+	   _("ai_bd_notice_cancel"), GTK_RESPONSE_CANCEL,
+	   NULL);
+      else
+	dialog = hildon_note_new_confirmation_add_buttons 
+	  (NULL,
+	   _("ai_ni_update_partly_successful"),
+	   _("ai_ni_bd_details"), 1,
+	   _("ai_bd_ok"), GTK_RESPONSE_CANCEL,
+	   NULL);
     }
   else
     {
-      if (c->cont)
-	c->cont (false, c->data);
-      delete c;
+      dialog = hildon_note_new_confirmation_add_buttons 
+	(NULL,
+	 _("ai_ni_update_list_not_successful"),
+	 _("ai_ni_bd_retry"), 2,
+	 _("ai_bd_notice_cancel"), GTK_RESPONSE_CANCEL,
+	 NULL);
+    }
+  
+  push_dialog (dialog);
+  g_signal_connect (dialog, "response",
+		    G_CALLBACK (rpc_error_response), c);
+  gtk_widget_show_all (dialog);
+}
+
+static void
+rpc_error_response (GtkWidget *dialog, gint response, gpointer data)
+{
+  rpc_clos *c = (rpc_clos *) data;
+
+  pop_dialog (dialog);
+  gtk_widget_destroy (dialog);
+
+  if (response == 1)
+    show_catalogue_dialog (c->catalogues, true,
+			   rpc_detailed_report_done, c);
+  else if (response == 2)
+    rpc_do_it (true, c);
+  else 
+    {
+      c->keep_going = (response == GTK_RESPONSE_OK
+		       && c->result_code != rescode_failure);
+      rpc_done (c);
     }
 }
 
-void
-refresh_package_cache_with_cont (int state,
-				 bool ask,
-				 void (*cont) (bool res, void *data), 
-				 void *data)
+static void
+rpc_detailed_report_done (bool changed, void *data)
 {
-  refreshed_this_session = true;
+  rpc_clos *c = (rpc_clos *) data;
 
-  rpc_clos *c = new rpc_clos;
-  c->cont = cont;
-  c->data = data;
-  c->state = state;
-
-  if (ask)
-    ask_yes_no (_("ai_nc_confirm_update"), refresh_package_cache_cont, c);
+  if (changed)
+    {
+      c->ask = true;
+      rpc_set_catalogues (c);
+    }
   else
-    refresh_package_cache_cont (true, c);
+    rpc_show_error_report (c);
+}
+
+static void
+rpc_done (void *data)
+{
+  rpc_clos *c = (rpc_clos *) data;
+
+  if (c->result_code == rescode_success
+      || c->result_code == rescode_partial_success)
+    {
+      last_update = time (NULL);
+      save_state ();
+      get_package_list_with_cont (c->state, rpc_end, c);
+    }
+  else
+    rpc_end (c);
+}
+
+static void
+rpc_end (void *data)
+{
+  rpc_clos *c = (rpc_clos *) data;
+
+  xexp_free (c->catalogues);
+      
+  c->cont (c->keep_going, c->data);
+  delete c;
+}
+
+/* Refreshing the package cache, wrapped in an interaction flow
+ */
+
+static void
+rpcf_end (bool success, void *unused)
+{
+  end_interaction_flow ();
 }
 
 void
-refresh_package_cache (int state, bool ask)
+refresh_package_cache_flow ()
 {
-  refresh_package_cache_with_cont (state, ask, NULL, NULL);
+  if (start_interaction_flow ())
+    refresh_package_cache (APTSTATE_DEFAULT,
+			   NULL, true, false,
+			   rpcf_end, NULL);
 }
 
 static int
@@ -1305,28 +1569,28 @@ maybe_refresh_package_cache ()
       && days_elapsed_since (last_update) < 30)
     return;
 
-  refresh_package_cache (APTSTATE_DEFAULT, true);
+  refresh_package_cache_flow ();
 }
 
 static void
 available_package_details (gpointer data)
 {
   package_info *pi = (package_info *)data;
-  show_package_details (pi, install_details, false, APTSTATE_DEFAULT);
+  show_package_details_flow (pi, install_details);
 }
 
 static void
-install_package_flow_end (void *data)
+install_package_flow_end (int n_successful, void *data)
 {
   package_info *pi = (package_info *)data;
   pi->unref ();
-  end_interaction_flow (GTK_WIDGET (get_main_window ()));
+  end_interaction_flow ();
 }
 
 static void
 install_package_flow (package_info *pi)
 {
-  if (start_interaction_flow (GTK_WIDGET (get_main_window ())))
+  if (start_interaction_flow ())
     {
       pi->ref ();
       install_package (pi, install_package_flow_end, pi);
@@ -1366,7 +1630,7 @@ static void
 installed_package_details (gpointer data)
 {
   package_info *pi = (package_info *)data;
-  show_package_details (pi, remove_details, false, APTSTATE_DEFAULT);
+  show_package_details_flow (pi, remove_details);
 }
 
 static void
@@ -1374,13 +1638,13 @@ uninstall_package_flow_end (void *data)
 {
   package_info *pi = (package_info *)data;
   pi->unref ();
-  end_interaction_flow (GTK_WIDGET (get_main_window ()));
+  end_interaction_flow ();
 }
 
 static void
 uninstall_package_flow (package_info *pi)
 {
-  if (start_interaction_flow (GTK_WIDGET (get_main_window ())))
+  if (start_interaction_flow ())
     {
       pi->ref ();
       uninstall_package (pi, uninstall_package_flow_end, pi);
@@ -1752,7 +2016,7 @@ search_packages_reply (int cmd, apt_proto_decoder *dec, void *data)
 
   if (!success)
     {
-      annoy_user_with_log (_("ai_ni_operation_failed"));
+      what_the_fock_p ();
       return;
     }
 
@@ -1775,8 +2039,19 @@ search_packages_reply (int cmd, apt_proto_decoder *dec, void *data)
       dec->decode_string_in_place (); // available_icon
 
       if (parent == &install_applications_view)
-	find_in_section_list (&result,
-			      install_sections, name);
+	{
+	  // We only search the first section in INSTALL_SECTIONS.
+	  // The first section in the list is either the special "All"
+	  // section that contains all packages, or there is only one
+	  // section.  In both cases, it is correct to only search the
+	  // first section.
+
+	  if (install_sections)
+	    {
+	      section_info *si = (section_info *)install_sections->data;
+	      find_in_package_list (&result, si->packages, name);
+	    }
+	}
       else if (parent == &upgrade_applications_view)
 	find_in_package_list (&result,
 			      upgradeable_packages, name);
@@ -1867,17 +2142,30 @@ search_packages (const char *pattern, bool in_descriptions)
     }
 }
 
+struct inp_clos {
+  void (*cont) (int n_successful, void *data);
+  void *data;
+};
+
+static void inp_no_package (void *data);
+static void inp_one_package (void *data);
+
 void
 install_named_package (int state, const char *package,
-		       void (*cont) (void *data), void *data)
+		       void (*cont) (int n_successful, void *data), void *data)
 {
   GList *p = NULL;
   GList *node = NULL;
 
   find_package_in_lists (state, &p, package);
 
+  inp_clos *c = new inp_clos;
+  c->cont = cont;
+  c->data = data;
+
   if (p == NULL)
-    annoy_user_with_cont (_("ai_ni_error_download_missing"), cont, data);
+    annoy_user (_("ai_ni_error_download_missing"),
+		inp_no_package, c);
   else
     {
       package_info *pi = (package_info *) p->data;
@@ -1885,12 +2173,15 @@ install_named_package (int state, const char *package,
 	{
 	  char *text = g_strdup_printf (_("ai_ni_package_installed"),
 					package);
-	  annoy_user_with_cont (text, cont, data);
+	  annoy_user (text, inp_one_package, c);
 	  pi->unref ();
 	  g_free (text);
 	}
       else
-	install_package (pi, cont, data);
+	{
+	  delete c;
+	  install_package (pi, cont, data);
+	}
 
       for (node = p->next; node != NULL; node = g_list_next (node))
 	{
@@ -1901,9 +2192,30 @@ install_named_package (int state, const char *package,
     }
 }
 
+static void
+inp_no_package (void *data)
+{
+  inp_clos *c = (inp_clos *)data;
+
+  c->cont (0, c->data);
+  delete c;
+}
+
+static void
+inp_one_package (void *data)
+{
+  inp_clos *c = (inp_clos *)data;
+
+  c->cont (1, c->data);
+  delete c;
+}
+
 void
-install_named_packages (int state, const char **packages, int install_type,
-			void (*cont) (void *data), void *data)
+install_named_packages (int state, const char **packages,
+			int install_type, bool automatic,
+			const char *title, const char *desc,
+			void (*cont) (int n_successful, void *data),
+			void *data)
 {
   GList *package_list = NULL;
   char **current_package = NULL;
@@ -1933,13 +2245,38 @@ install_named_packages (int state, const char **packages, int install_type,
 	    }
 	}
       else
-	add_log ("Package %s not found\n", *current_package);
+	{
+	  /* Create a 'fake' package_info structure so that we at
+	     least have something to display.
+	  */
+	  package_info *pi = new package_info;
+	  pi->name = g_strdup (*current_package);
+	  pi->available_version = g_strdup ("");
+
+	  pi->have_info = true;
+	  pi->info.installable_status = status_not_found;
+	  pi->info.download_size = 0;
+	  pi->info.install_user_size_delta = 0;
+	  pi->info.required_free_space = 0;
+	  pi->info.install_flags = 0;
+	  pi->info.removable_status = status_able;
+	  pi->info.remove_user_size_delta = 0;
+
+	  pi->have_detail_kind = install_details;
+	  pi->summary = g_strdup_printf (_("ai_va_details_unable_install"),
+					 pi->name);
+	  pi->summary_packages[sumtype_missing] =
+	    g_list_append (NULL, g_strdup (pi->name));
+
+	  package_list = g_list_append (package_list, pi);
+	}
 
       g_list_free (search_list);
     }
   
   install_packages (package_list,
-		    state, install_type,
+		    state, install_type, automatic,
+		    title, desc,
 		    cont, data);
 }
 
@@ -2008,29 +2345,26 @@ set_operation_callback (void (*func) (gpointer), gpointer data)
  */
 
 static void rp_restore (bool res, void *data);
-static void rp_end (void *data);
+static void rp_unsuccessful (void *data);
+static void rp_end (int n_successful, void *data);
 
 void
 restore_packages_flow ()
 {
-  if (start_interaction_flow (GTK_WIDGET (get_main_window ())))
+  if (start_interaction_flow ())
     {
       char *filename =
-	g_strdup_printf ("%s/.hildon-application-manager.backup",
-			 g_get_home_dir ());
+	g_strdup_printf ("%s/%s",
+			 g_get_home_dir (), RESTORE_BACKUP_FILENAME);
       xexp *backup = xexp_read_file (filename);
       g_free (filename);
       
       if (backup)
-	{
-	  refresh_package_cache_with_cont (APTSTATE_DEFAULT, false,
-					   rp_restore, backup);
-	}
+	refresh_package_cache (APTSTATE_DEFAULT,
+			       NULL, false, true,
+			       rp_restore, backup);
       else
-	{
-	  annoy_user_with_cont (_("ai_ni_operation_failed"),
-				rp_end, backup);
-	}
+	annoy_user (_("ai_ni_operation_failed"), rp_unsuccessful, backup);
     }
 }
 
@@ -2039,32 +2373,45 @@ rp_restore (bool res, void *data)
 {
   xexp *backup = (xexp *)data;
 
-  int len = xexp_length (backup);
-  const char **names = (const char **)new char* [len+1];
-
-  xexp *p = xexp_first (backup);
-  int i = 0;
-  while (p)
+  if (res)
     {
-      if (xexp_is (p, "pkg") && xexp_is_text (p))
-	names[i++] = xexp_text (p);
-      p = xexp_rest (p);
+      int len = xexp_length (backup);
+      const char **names = (const char **)new char* [len+1];
+      
+      xexp *p = xexp_first (backup);
+      int i = 0;
+      while (p)
+	{
+	  if (xexp_is (p, "pkg") && xexp_is_text (p))
+	    names[i++] = xexp_text (p);
+	  p = xexp_rest (p);
+	}
+      names[i] = NULL;
+      install_named_packages (APTSTATE_DEFAULT, names,
+			      INSTALL_TYPE_BACKUP, false,
+			      NULL, NULL,
+			      rp_end, backup);
+      delete names;
     }
-  names[i] = NULL;
-  install_named_packages (APTSTATE_DEFAULT, names, INSTALL_TYPE_BACKUP,
-			  rp_end, backup);
-  delete names;
+  else
+    rp_end (0, backup);
 }
 
 static void
-rp_end (void *data)
+rp_unsuccessful (void *data)
+{
+  rp_end (0, data);
+}
+
+static void
+rp_end (int n_successful, void *data)
 {
   xexp *backup = (xexp *)data;
 
   if (backup)
     xexp_free (backup);
 
-  end_interaction_flow (GTK_WIDGET (get_main_window ()));
+  end_interaction_flow ();
 }
 
 /* Updating all installed system-update packages
@@ -2073,22 +2420,26 @@ rp_end (void *data)
 static void us_get_system_packages (bool res, void *data);
 static void us_get_system_packages_reply (int cmd, apt_proto_decoder *dec,
 					  void *data);
-static void us_end (void *data);
+static void us_end (int n_successful, void *data);
 
 void
 update_system_flow ()
 {
-  if (start_interaction_flow (GTK_WIDGET (get_main_window ())))
-    refresh_package_cache_with_cont (APTSTATE_DEFAULT, FALSE,
-				     us_get_system_packages, NULL);
+  if (start_interaction_flow ())
+    refresh_package_cache (APTSTATE_DEFAULT,
+			   NULL, false, true,
+			   us_get_system_packages, NULL);
 }
 
 static void
 us_get_system_packages (bool res, void *data)
 {
-  apt_worker_get_system_update_packages (APTSTATE_DEFAULT,
-					 us_get_system_packages_reply,
-					 NULL);
+  if (res)
+    apt_worker_get_system_update_packages (APTSTATE_DEFAULT,
+					   us_get_system_packages_reply,
+					   data);
+  else
+    us_end (0, data);
 }
 
 static void
@@ -2098,7 +2449,7 @@ us_get_system_packages_reply (int cmd, apt_proto_decoder *dec, void *data)
 
   if (dec == NULL)
     {
-      us_end (data);
+      us_end (0, data);
       return;
     }
 
@@ -2113,71 +2464,67 @@ us_get_system_packages_reply (int cmd, apt_proto_decoder *dec, void *data)
     }
   packages[i] = NULL;
 
-  install_named_packages (APTSTATE_DEFAULT,
-			  (const char**)packages, INSTALL_TYPE_UPDATE_SYSTEM,
+  install_named_packages (APTSTATE_DEFAULT, (const char**)packages,
+			  INSTALL_TYPE_UPDATE_SYSTEM, false,
+			  NULL, NULL,
 			  us_end, data);
 }
 
 static void
-us_end (void *data)
+us_end (int n_succesful, void *data)
 {
-  end_interaction_flow (GTK_WIDGET (get_main_window ()));
+  end_interaction_flow ();
 }
 
-/* Installing from file
+/* INSTALL_FROM_FILE_FLOW
  */
 
-static void
-install_from_file_flow_end (void *data)
-{
-  char *filename = (char *)data;
-  cleanup_temp_file ();
-  end_interaction_flow (GTK_WIDGET (get_main_window ()));
-  g_free (filename);
-}
-
-static void
-install_from_local_file (char *filename, void *unused)
-{
-  if (filename)
-    {
-      if (g_str_has_suffix (filename, ".install"))
-	open_local_install_instructions (filename,
-					 install_from_file_flow_end,
-					 filename);
-      else
-	install_local_deb_file (filename,
-				install_from_file_flow_end, filename);
-    }
-  else
-    install_from_file_flow_end (filename);
-}
-
-static void
-install_from_file_cont (char *uri, void *unused)
-{
-  localize_file_and_keep_it_open (uri, install_from_local_file, NULL);
-  g_free (uri);
-}
+static void iff_with_filename (char *uri, void *unused);
+static void iff_end (bool success, void *unused);
 
 void
 install_from_file_flow (const char *filename)
 {
-  if (start_interaction_flow (GTK_WIDGET (get_main_window ())))
+  if (start_interaction_flow ())
     {
       if (filename == NULL)
-	show_deb_file_chooser (install_from_file_cont, NULL);
+	show_deb_file_chooser (iff_with_filename, NULL);
       else
-	install_from_file_cont (g_strdup (filename), NULL);
+	iff_with_filename (g_strdup (filename), NULL);
     }
 }
 
+static void
+iff_with_filename (char *uri, void *unused)
+{
+  if (uri)
+    {
+      install_file (uri, iff_end, NULL);
+      g_free (uri);
+    }
+  else
+    iff_end (false, NULL);
+}
+
+static void
+iff_end (bool success, void *unused)
+{
+  end_interaction_flow ();
+}
+
 static void
 window_destroy (GtkWidget* widget, gpointer data)
 {
   gtk_main_quit ();
 }
 
+static gboolean
+window_delete_event (GtkWidget* widget, GdkEvent *ev, gpointer data)
+{
+  menu_close ();
+  return TRUE;
+}
+
 static void
 main_window_realized (GtkWidget* widget, gpointer unused)
 {
@@ -2191,6 +2538,12 @@ main_window_realized (GtkWidget* widget, gpointer unused)
 	      "hildon-application-manager");
 }
 
+static void
+cancel_download (void *unused)
+{
+  cancel_apt_worker ();
+}
+
 static void
 apt_status_callback (int cmd, apt_proto_decoder *dec, void *unused)
 {
@@ -2202,28 +2555,22 @@ apt_status_callback (int cmd, apt_proto_decoder *dec, void *unused)
   int total = dec->decode_int ();
 
   if (total > 0)
-    set_progress ((apt_proto_operation)op, already, total);
+    {
+      if (op == op_downloading)
+	{
+	  set_entertainment_download_fun (already, total);
+	  set_entertainment_cancel (cancel_download, NULL);
+	}
+      else
+	{
+	  set_entertainment_fun (NULL, already, total);
+	  set_entertainment_cancel (NULL, NULL);
+	}
+    }
 }
 
 static GtkWindow *main_window = NULL;
 
-static void
-mime_open_handler (gpointer raw_data, int argc, char **argv)
-{
-  if (argc > 0)
-    {
-      const char *filename = argv[0];
-
-      present_main_window ();
-      if (strcmp (filename, "magic:restore-packages") == 0)
-	restore_packages_flow ();
-      if (strcmp (filename, "magic:update-system") == 0)
-	update_system_flow ();
-      else
-	install_from_file_flow (filename);
-    }
-}
-
 static GtkWidget *toolbar_operation_label = NULL;
 static GtkWidget *toolbar_operation_item = NULL;
 
@@ -2258,13 +2605,6 @@ get_device_label ()
   return device_label;
 }
 
-void
-present_main_window ()
-{
-  if (main_window)
-    gtk_window_present (main_window);
-}
-
 static void
 set_current_toolbar_visibility (bool f)
 {
@@ -2307,7 +2647,7 @@ set_toolbar_visibility (bool fullscreen, bool visibility)
       if (!is_fullscreen)
 	set_current_toolbar_visibility (visibility);
     }
-  save_settings ();
+  save_state ();
 }
 
 static gboolean
@@ -2424,7 +2764,7 @@ set_current_help_topic (const char *topic)
 static void
 call_refresh_package_cache (GtkWidget *button, gpointer data)
 {
-  refresh_package_cache (APTSTATE_DEFAULT, true);
+  refresh_package_cache_flow ();
 }
 
 /* Take a snapshot of the data we want to keep in a backup.
@@ -2445,22 +2785,78 @@ save_backup_data_reply (int cmd, apt_proto_decoder *dec, void *data)
   /* No action required */
 }
 
+struct pending_cont {
+  pending_cont *next;
+  void (*cont) (void *data);
+  void *data;
+};
+
+static bool initial_packages_available = false;
+static pending_cont *pending_for_initial_packages = NULL;
+
+static void
+notice_initial_packages_available (gpointer data)
+{
+  initial_packages_available = true;
+
+  while (pending_for_initial_packages)
+    {
+      pending_cont *c = pending_for_initial_packages;
+      pending_for_initial_packages = c->next;
+      c->cont (c->data);
+      delete c;
+    }
+}
+
+void
+with_initialized_packages (void (*cont) (void *data), void *data)
+{
+  if (initial_packages_available)
+    cont (data);
+  else
+    {
+      pending_cont *c = new pending_cont;
+      c->cont = cont;
+      c->data = data;
+      c->next = pending_for_initial_packages;
+      pending_for_initial_packages = c;
+    }
+}
+
 int
 main (int argc, char **argv)
 {
   GtkWidget *window;
   GtkWidget *toolbar, *image;
-  GtkMenu *main_menu;
   char *apt_worker_prog = "/usr/libexec/apt-worker";
+  bool show = true;
+
+  if (argc > 1 && !strcmp (argv[1], "--no-show"))
+    {
+      show = false;
+      argc--;
+      argv++;
+    }
+  if (argc > 1)
+    {
+      apt_worker_prog = argv[1];
+      argc--;
+      argv++;
+    }
 
   setlocale (LC_ALL, "");
-  bind_textdomain_codeset ("osso-application-installer", "UTF-8");
-  textdomain ("osso-application-installer");
+  bind_textdomain_codeset ("hildon-application-manager", "UTF-8");
+  textdomain ("hildon-application-manager");
 
   load_settings ();
+  load_state ();
 
   gtk_init (&argc, &argv);
-  setup_dbus();
+
+  init_dbus_or_die (show);
+
+  osso_ctxt = osso_initialize ("hildon_application_manager",
+			       PACKAGE_VERSION, TRUE, NULL);
 
   // XXX - We don't want a two-part title and this seems to be the
   //       only way to get rid of it.  Hopefully, setting an empty
@@ -2470,16 +2866,6 @@ main (int argc, char **argv)
 
   clear_log ();
 
-  // XXX - stupid option parsing, should be improved but options are
-  //       only used for testing right now.
-  //
-  if (argc > 1)
-    {
-      apt_worker_prog = argv[1];
-      argc--;
-      argv++;
-    }
-
   window = hildon_window_new ();
   gtk_window_set_title (GTK_WINDOW (window), _("ai_ap_application_installer"));
 
@@ -2535,7 +2921,7 @@ main (int argc, char **argv)
   gtk_tool_item_set_expand (GTK_TOOL_ITEM (search_button), TRUE);
   gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (search_button), TRUE);
   g_signal_connect (search_button, "clicked",
-		    G_CALLBACK (show_search_dialog),
+		    G_CALLBACK (show_search_dialog_flow),
 		    NULL);
   g_signal_connect (G_OBJECT (search_button), "insensitive_press",
 		    G_CALLBACK (insensitive_press),
@@ -2571,36 +2957,32 @@ main (int argc, char **argv)
 
   gtk_box_pack_start (GTK_BOX (main_vbox), main_trail, FALSE, FALSE, 0);
 
+  g_signal_connect (G_OBJECT (window), "delete-event",
+		    G_CALLBACK (window_delete_event), NULL);
+
   g_signal_connect (G_OBJECT (window), "destroy",
 		    G_CALLBACK (window_destroy), NULL);
 
   g_signal_connect (G_OBJECT (window), "realize",
 		    G_CALLBACK (main_window_realized), NULL);
 
-  main_menu = GTK_MENU (gtk_menu_new ());
-  create_menu (main_menu);
-  hildon_window_set_menu (HILDON_WINDOW (window), GTK_MENU (main_menu));
+  create_menu (HILDON_WINDOW (window));
 
   show_view (&main_view);
   set_toolbar_visibility (true, fullscreen_toolbar);
   set_toolbar_visibility (false, normal_toolbar);
 
   if (!start_apt_worker (apt_worker_prog))
-    annoy_user (_("ai_ni_operation_failed"));
+    what_the_fock_p ();
 
   apt_worker_set_status_callback (apt_status_callback, NULL);
 
-  get_package_list (APTSTATE_DEFAULT);
+  get_package_list_with_cont (APTSTATE_DEFAULT,
+			      notice_initial_packages_available, NULL);
   save_backup_data ();
 
-  /* XXX - check errors.
-   */
-  osso_ctxt = osso_initialize ("hildon_application_manager",
-			       PACKAGE_VERSION, TRUE, NULL);
-
-  osso_mime_set_cb (osso_ctxt, mime_open_handler, NULL);
-
-  gtk_widget_show_all (window);
+  if (show)
+    present_main_window ();
 
   gtk_main ();
 }
