@@ -81,6 +81,12 @@ void get_intermediate_package_info (package_info *pi,
 				    void *,
 				    int state);
 
+void get_intermediate_package_list_info (GList *packages,
+					 bool only_installable_info,
+					 void (*cont) (void *data),
+					 void *data,
+					 int state);
+
 struct section_info {
 
   section_info ();
@@ -104,19 +110,44 @@ void show_current_details ();
 void do_current_operation ();
 
 void install_named_package (int state, const char *package,
-			    void (*cont) (void *data), void *data);
-void install_named_packages (int state, const char **package, int install_type,
-			     void (*cont) (void *data), void *data);
+			    void (*cont) (int n_successful, void *data),
+			    void *data);
+void install_named_packages (int state, const char **packages,
+			     int install_type, bool automatic,
+			     const char *title, const char *desc,
+			     void (*cont) (int n_successful, void *data),
+			     void *data);
 
-void refresh_package_cache (int state, 
-			    bool ask);
-void refresh_package_cache_with_cont (int state,
-				      bool ask,
-				      void (*cont) (bool res, void *data), 
-				      void *data);
+/* REFRESH_PACKAGE_CACHE will download and digest the indices for all
+   enabled catalogues.  It will show error messages to the user and
+   allow the catalogue configuration to be edited and the operation to
+   be retried in case of errors.
+
+   When CATALOGUES is non-NULL, it will replace the existing catalogue
+   configuration before the operation begins.
+
+   When ASK is true, the user will be asked whether to actually
+   perform the operation.
+
+   When CONTINUED is true, error messages will include "Continue" and
+   "Stop" buttons instead of "Ok".
+
+   CONT is called when the operation has finished.  KEEP_GOING
+   indicates whether the user wants to continue or not in case an
+   error occurred.  When CONTINUED is false, KEEP_GOING will be true
+   when there was no fatal error and continuing might make sense.
+*/
+void refresh_package_cache (int state,
+			    xexp *new_catalogues,
+			    bool ask, bool continued,
+			    void (*cont) (bool keep_going, void *data), 
+			    void *data);
+
+void refresh_package_cache_flow ();
 
 void install_from_file_flow (const char *filename);
 void restore_packages_flow ();
+void update_system_flow ();
 
 void save_backup_data ();
 
@@ -134,7 +165,6 @@ GtkWidget *get_device_label ();
 
 void set_fullscreen (bool);
 void toggle_fullscreen ();
-void present_main_window ();
 
 void set_toolbar_visibility (bool fullscreen, bool visibility);
 
@@ -143,4 +173,6 @@ void set_toolbar_visibility (bool fullscreen, bool visibility);
 void set_dialog_help (GtkWidget *dialog, const char *topic);
 void show_help ();
 
+void with_initialized_packages (void (*cont) (void *data), void *data);
+
 #endif /* !MAIN_H */
