@@ -31,6 +31,7 @@
 #include <hildon/hildon-note.h>
 
 #include "operations.h"
+#include "instr.h"
 #include "util.h"
 #include "main.h"
 #include "apt-worker-client.h"
@@ -40,6 +41,8 @@
 
 #define _(x) gettext (x)
 
+#define HAM_BACKUP_RESPONSE 1
+
 /* Common utilities
  */
 
@@ -69,39 +72,54 @@ void
 installable_status_to_message (package_info *pi,
 			       char *&msg, bool &with_details)
 {
-  const char *smsg = NULL;
-
   if (pi->info.installable_status == status_missing)
     {
-      smsg = (pi->installed_version
-	      ? _("ai_ni_error_update_missing")
-	      : _("ai_ni_error_install_missing"));
+      msg = g_strdup_printf ("%s\n%s",
+			     (pi->installed_version
+			      ? _("ai_ni_error_update_missing")
+			      : _("ai_ni_error_install_missing")),
+			     pi->get_display_name (false));
       with_details = true;
     }
   else if (pi->info.installable_status == status_conflicting)
     {
-      smsg = (pi->installed_version
-	      ? _("ai_ni_error_update_conflict")
-	      : _("ai_ni_error_install_conflict"));
+      msg = g_strdup_printf ("%s\n%s",
+			     (pi->installed_version
+			      ? _("ai_ni_error_update_conflict")
+			      : _("ai_ni_error_install_conflict")),
+			     pi->get_display_name (false));
       with_details = true;
     }
   else if (pi->info.installable_status == status_corrupted)
     {
-      smsg = (pi->installed_version
-	      ? _("ai_ni_error_update_corrupted")
-	      : _("ai_ni_error_install_corrupted"));
+      msg = g_strdup_printf ("%s\n%s",
+			     (pi->installed_version
+			      ? _("ai_ni_error_update_corrupted")
+			      : _("ai_ni_error_install_corrupted")),
+			     pi->get_display_name (false));
       with_details = false;
     }
   else if (pi->info.installable_status == status_incompatible)
     {
-      smsg = (pi->installed_version
-	      ? _("ai_ni_error_update_incompatible")
-	      : _("ai_ni_error_install_incompatible"));
+      msg = g_strdup_printf ("%s\n%s",
+			     (pi->installed_version
+			      ? _("ai_ni_error_update_incompatible")
+			      : _("ai_ni_error_install_incompatible")),
+			     pi->get_display_name (false));
       with_details = false;
     }
   else if (pi->info.installable_status == status_incompatible_current)
     {
-      smsg = _("ai_ni_error_n770package_incompatible");
+      msg = g_strdup_printf ("%s\n%s",
+			     _("ai_ni_error_n770package_incompatible"),
+			     pi->get_display_name (false));
+      with_details = false;
+    }
+  else if (pi->info.installable_status == status_not_found)
+    {
+      msg = g_strdup_printf ("%s\n%s",
+			     _("ai_ni_error_download_missing"),
+			     pi->get_display_name (false));
       with_details = false;
     }
   else
@@ -112,9 +130,6 @@ installable_status_to_message (package_info *pi,
 			     pi->get_display_name (false));
       with_details = true;
     }
-
-  if (smsg)
-    msg = g_strdup (smsg);
 }
 
 /* INSTALL_PACKAGES - Overview
@@ -183,6 +198,7 @@ installable_status_to_message (package_info *pi,
 struct ip_clos {
   int install_type;
   int state;
+  bool automatic;
 
   GList *all_packages;   // all packages given to install_packages
   GList *packages;       // the ones that are not installed or uptodate
@@ -195,12 +211,15 @@ struct ip_clos {
   GSList *upgrade_versions; // to upgrade to.
 
   // at the end
+  bool entertaining;        // is the progress bar up?
+  int n_successful;         // how many have been installed successfully
   bool reboot;              // whether to reboot
   
-  void (*cont) (void *);
+  void (*cont) (int n_successful, void *);
   void *data;
 };
 
+static void ip_confirm_with_info (package_info *pi, void *data, bool changed);
 static void ip_confirm_install_response (bool res, void *data);
 static void ip_select_package_response (gboolean res, GList *selected_packages,
 					void *data);
@@ -224,8 +243,13 @@ static void ip_install_cur_reply (int cmd, apt_proto_decoder *dec, void *data);
 static void ip_clean_reply (int cmd, apt_proto_decoder *dec, void *data);
 static void ip_install_next (void *data);
 
+static gboolean ip_suggest_backup (ip_clos *c);
+static void ip_suggest_backup_cmd_done (int status, void *data);
+static void ip_close_apps (bool res, void *data);
+
 static void ip_show_cur_details (void *data);
 static void ip_show_cur_problem_details (void *data);
+static void ip_show_details_done (void *data);
 
 static void ip_abort_cur (ip_clos *c, const char *msg, bool with_details);
 static void ip_abort_cur_with_status_details (ip_clos *c);
@@ -233,32 +257,40 @@ static void ip_abort_response (GtkDialog *dialog, gint response,
 			       gpointer data);
 
 static void ip_end (void *data);
+static void ip_end_after_reboot (void *data);
+static void ip_end_rebooting (void *data);
 
 void
 install_package (package_info *pi,
-		 void (*cont) (void *), void *data)
+		 void (*cont) (int n_successful, void *), void *data)
 {
   install_packages (g_list_prepend (NULL, pi),
-		    APTSTATE_DEFAULT, INSTALL_TYPE_STANDARD,
+		    APTSTATE_DEFAULT, INSTALL_TYPE_STANDARD, false,
+		    NULL, NULL,
 		    cont, data);
 }
 
 void
 install_packages (GList *packages,
 		  int state, int install_type,
-		  void (*cont) (void *), void *data)
+		  bool automatic,
+		  const char *title, const char *desc,
+		  void (*cont) (int n_successful, void *), void *data)
 {
   ip_clos *c = new ip_clos;
 
   c->install_type = install_type;
   c->state = state;
+  c->automatic = automatic;
   c->all_packages = packages;
   c->cont = cont;
   c->data = data;
   c->upgrade_names = NULL;
   c->upgrade_versions = NULL;
+  c->n_successful = 0;
   c->reboot = false;
-  
+  c->entertaining = false;
+
   // Filter packages, stopping after the first when this is a standard
   // install.
 
@@ -276,59 +308,88 @@ install_packages (GList *packages,
 
   if (c->packages == NULL)
     {
-      annoy_user_with_cont ("Nothing to install", ip_end, c);  // XXX-L10N
+      annoy_user (_("ai_ni_all_installed"), ip_end, c);
       return;
     }
 
   // Bring up the appropriate confirmation dialog.
 
   if (c->install_type == INSTALL_TYPE_BACKUP
-      || c->install_type == INSTALL_TYPE_MEMORY_CARD)
+      || c->install_type == INSTALL_TYPE_MEMORY_CARD
+      || c->install_type == INSTALL_TYPE_MULTI)
     {
-      select_package_list (c->packages, 
-			   _("ai_ti_install_apps"), 
-			   (c->install_type == INSTALL_TYPE_BACKUP
-			    ? _("ai_ti_restore")
-			    : _("ai_ti_memory")),
+      if (title == NULL)
+	title = _("ai_ti_install_apps");
+
+      if (desc == NULL)
+	{
+	  switch (c->install_type)
+	    {
+	    case INSTALL_TYPE_BACKUP:
+	      desc = _("ai_ia_restore");
+	      break;
+	    case INSTALL_TYPE_MEMORY_CARD:
+	      desc = _("ai_ia_memory");
+	      break;
+	    default:
+	      desc = _("Install");
+	      break;
+	    }
+	}
+
+      select_package_list (c->packages,
+			   c->state,
+			   title, desc,
 			   ip_select_package_response, c);
     }
   else if (c->install_type != INSTALL_TYPE_UPDATE_SYSTEM)
     {
-      GString *text = g_string_new ("");
-      char download_buf[20];
       package_info *pi;
-
+      
       c->cur = c->packages;
       pi = (package_info *)(c->cur->data);
 
-      size_string_general (download_buf, 20, pi->info.download_size);
-
-      g_string_printf (text,
-		       (pi->installed_version
-			? _("ai_nc_update")
-			: _("ai_nc_install")),
-		       pi->get_display_name (false),
-		       pi->available_version, download_buf);
-  
-      ask_yes_no_with_arbitrary_details ((pi->installed_version
-					  ? _("ai_ti_confirm_update")
-					  : _("ai_ti_confirm_install")),
-					 text->str,
-					 ip_confirm_install_response,
-					 ip_show_cur_details, c);
-      g_string_free (text, 1);
+      get_intermediate_package_info (pi, true,
+				     ip_confirm_with_info, c, c->state);
     }
   else
     ip_confirm_install_response (true, c);
 }
 
+static void
+ip_confirm_with_info (package_info *pi, void *data, bool changed)
+{
+  ip_clos *c = (ip_clos *)data;
+
+  GString *text = g_string_new ("");
+  char download_buf[20];
+  
+  size_string_general (download_buf, 20, pi->info.download_size);
+
+  g_string_printf (text,
+		   (pi->installed_version
+		    ? _("ai_nc_update")
+		    : _("ai_nc_install")),
+		   pi->get_display_name (false),
+		   pi->available_version, download_buf);
+  
+  ask_yes_no_with_arbitrary_details ((pi->installed_version
+				      ? _("ai_ti_confirm_update")
+				      : _("ai_ti_confirm_install")),
+				     text->str,
+				     ip_confirm_install_response,
+				     ip_show_cur_details, c);
+  g_string_free (text, 1);
+}
+
 static void
 ip_show_cur_details (void *data)
 {
   ip_clos *c = (ip_clos *)data;
   package_info *pi = (package_info *)(c->cur->data);
   
-  show_package_details (pi, install_details, false, c->state);
+  show_package_details (pi, install_details, false, c->state,
+			ip_show_details_done, c);
 }
 
 static void
@@ -337,7 +398,13 @@ ip_show_cur_problem_details (void *data)
   ip_clos *c = (ip_clos *)data;
   package_info *pi = (package_info *)(c->cur->data);
   
-  show_package_details (pi, install_details, true, c->state);
+  show_package_details (pi, install_details, true, c->state,
+			ip_show_details_done, c);
+}
+
+static void
+ip_show_details_done (void *data)
+{
 }
 
 static void
@@ -346,8 +413,19 @@ ip_select_package_response (gboolean res, GList *selected_packages,
 {
   ip_clos *c = (ip_clos *)data;
 
-  if (!res || selected_packages == NULL)
-    ip_end (c);
+  if (!res)
+    {
+      if (c->install_type == INSTALL_TYPE_MEMORY_CARD
+	  && c->automatic)
+	annoy_user (_("ai_ni_memory_cancelled"),
+		    ip_end, c);
+      else
+	ip_end (c);
+    }
+  else if (selected_packages == NULL)
+    {
+      ip_end (c);
+    }
   else
     {
       g_list_free (c->packages);
@@ -370,6 +448,17 @@ ip_confirm_install_response (bool res, void *data)
 static void
 ip_ensure_network (ip_clos *c)
 {
+  /* Start entertaining the user here.  We stop in ip_end, at the
+     last.
+   */
+
+  set_entertainment_fun (NULL, -1, 0);
+  set_entertainment_cancel (NULL, NULL);
+  set_entertainment_title ("");
+  start_entertaining_user ();
+  
+  c->entertaining = true;
+
   if (c->install_type != INSTALL_TYPE_MEMORY_CARD)
     ensure_network (ip_ensure_network_reply, c);
   else
@@ -401,8 +490,18 @@ ip_check_cert_loop (ip_clos *c)
     {
       package_info *pi = (package_info *)c->cur->data;
       printf ("CHECK CERT: %s\n", pi->name);
-      apt_worker_install_check (c->state, pi->name,
-				ip_check_cert_reply, c);
+
+      if (pi->have_info 
+	  && pi->info.installable_status == status_not_found)
+	{
+	  /* Skip packages that we know don't exist.
+	   */
+	  c->cur = c->cur->next;
+	  ip_check_cert_loop (c);
+	}
+      else
+	apt_worker_install_check (c->state, pi->name,
+				  ip_check_cert_reply, c);
     }
   else
     {
@@ -450,12 +549,13 @@ ip_check_cert_reply (int cmd, apt_proto_decoder *dec, void *data)
   if (some_domains_violated)
     {
       if (red_pill_mode)
-	ask_custom (_("Trusted upgrade path broken"),
-		    "Continue anyway", "Stop",
+	ask_custom (_("ai_ni_error_broken_path"),
+		    _("ai_bd_continue"),
+		    _("ai_bd_cancel"),
 		    ip_legalese_response, c);
       else
-	annoy_user_with_cont (_("Trusted upgrade path broken"),
-			      ip_end, c);
+	annoy_user (_("ai_ni_error_broken_path"),
+		    ip_end, c);
     }
   else if (some_not_certfied)
     scare_user_with_legalese (false, ip_legalese_response, c);
@@ -497,7 +597,49 @@ static void
 ip_install_loop (ip_clos *c)
 {
   if (c->cur == NULL)
-    ip_end (c);
+    {
+      /* End of loop, show a success report to the user.
+
+         If there is only one package in the list, talk specifically
+         about that package.  Otherwise, just show the number of
+         packages that have been successfully handled.
+       */
+
+      stop_entertaining_user ();
+      c->entertaining = false;
+
+      if (c->n_successful > 0)
+	{
+	  if (c->reboot)
+	    {
+	      /* ip_end will show the 'success' dialog for us.
+	       */
+	      ip_end (c);
+	    }
+	  else if (c->all_packages->next == NULL)
+	    {
+	      package_info *pi = (package_info *)c->all_packages->data;
+	      char *str = g_strdup_printf ((pi->installed_version != NULL
+					    ? _("ai_ni_update_successful")
+					    : _("ai_ni_install_successful")),
+					   pi->get_display_name (false));
+	      annoy_user (str, ip_end, c);
+	      g_free (str);
+	    }
+	  else
+	    {
+	      char *str =
+		g_strdup_printf (ngettext ("ai_ni_multiple_install",
+					   "ai_ni_multiple_installs", 
+					   c->n_successful),
+				 c->n_successful);
+	      annoy_user (str, ip_end, c);
+	      g_free (str);
+	    }
+	}
+      else
+	ip_end (c);
+    }
   else
     {
       package_info *pi = (package_info *)(c->cur->data);
@@ -513,6 +655,15 @@ ip_install_with_info (package_info *pi, void *data, bool changed)
   
   if (pi->info.installable_status == status_able)
     {
+      if ((pi->info.install_flags & pkgflag_suggest_backup) &&
+	  !ip_suggest_backup (c))
+	{
+	  /* if suggest_flag is present and user chooses 'Cancel'
+	     option, installation for this package is aborted */
+	  ip_end (c);
+	  return;
+	}
+
       add_log ("-----\n");
       if (pi->installed_version)
 	add_log ("Upgrading %s %s to %s\n", pi->name,
@@ -525,14 +676,6 @@ ip_install_with_info (package_info *pi, void *data, bool changed)
 	annoy_user_with_errno (errno, "get_free_space",
 			       ip_end, c);
 
-      /* XXX - What is the right thing to do when a packages that asks
-    	       for a reboot fails during installation?  Is it better
-    	       to reboot or not?
-      */
-
-      if (pi->info.install_flags & pkgflag_reboot)
-	c->reboot = true;
-
       if (pi->info.required_free_space + pi->info.download_size >= free_space)
 	{
 	  char free_string[20];
@@ -549,9 +692,7 @@ ip_install_with_info (package_info *pi, void *data, bool changed)
 	}
       else if (pi->info.install_flags & pkgflag_close_apps)
 	{
-	  // XXX - L10N
-	  annoy_user_with_cont (_("You should close all Applications now."),
-				ip_install_cur, c);
+	  ask_yes_no (_("ai_nc_close_apps"), ip_close_apps, c);
 	}
       else
 	{
@@ -612,7 +753,7 @@ ip_check_upgrade_reply (int cmd, apt_proto_decoder *dec, void *data)
   if (success)
     ip_check_upgrade_loop (c);
   else
-    annoy_user_with_cont (_("ai_ni_operation_failed"), ip_end, c);
+    annoy_user (_("ai_ni_operation_failed"), ip_end, c);
 }
 
 static void
@@ -679,7 +820,15 @@ ip_install_cur (void *data)
   package_info *pi = (package_info *)(c->cur->data);
 
   printf ("INSTALL %s\n", pi->name);
-  
+
+  char *title = g_strdup_printf (pi->installed_version
+				 ? _("ai_nw_updating")
+				 : _("ai_nw_installing"),
+				 pi->get_display_name (false));
+  set_entertainment_title (title);
+  set_entertainment_fun (NULL, -1, 0);
+  g_free (title);
+
   set_log_start ();
   apt_worker_install_package (c->state, pi->name,
 			      pi->installed_version != NULL,
@@ -692,8 +841,6 @@ ip_install_cur_reply (int cmd, apt_proto_decoder *dec, void *data)
   ip_clos *c = (ip_clos *)data;
   package_info *pi = (package_info *)(c->cur->data);
 
-  hide_progress ();
-
   if (dec == NULL)
     {
       ip_end (c);
@@ -708,21 +855,16 @@ ip_install_cur_reply (int cmd, apt_proto_decoder *dec, void *data)
 
   if (result_code == rescode_success)
     {
-      if (c->install_type == INSTALL_TYPE_STANDARD)
-	{
-	  char *str = g_strdup_printf ((pi->installed_version != NULL
-					? _("ai_ni_update_successful")
-					: _("ai_ni_install_successful")),
-				       pi->get_display_name (false));
-	  annoy_user_with_cont (str, ip_install_next, c);
-	  g_free (str);
-	}
-      else
-	ip_install_next (c);
+      c->n_successful += 1;
+
+      if (pi->info.install_flags & pkgflag_reboot)
+	c->reboot = true;
+
+      ip_install_next (c);
     }
   else
     {
-      if (progress_was_cancelled ())
+      if (entertainment_was_cancelled ())
 	ip_end (c);
       else
 	{
@@ -760,6 +902,93 @@ ip_install_next (void *data)
   ip_install_loop (c);
 }
 
+static gboolean
+ip_suggest_backup (ip_clos *c)
+{
+  GtkWidget *dialog = NULL;
+  gint result = G_MAXINT; 
+  gboolean keep_asking = TRUE;
+  
+  /* Show custom dialog if suggest_backup flag is present */
+  dialog = gtk_dialog_new_with_buttons
+    (_("ai_ti_create_backup"),
+     NULL,
+     GTK_DIALOG_MODAL,
+     _("ai_bd_confirm_ok"), GTK_RESPONSE_OK,
+     _("ai_bd_backup"), HAM_BACKUP_RESPONSE,
+     _("ai_bd_confirm_cancel"), GTK_RESPONSE_CANCEL,
+     NULL);
+  push_dialog (dialog);
+	  
+  gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
+  GtkWidget *label = gtk_label_new (_("ai_ia_backup"));
+  gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
+  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox),
+		     label);	  
+  gtk_widget_show_all (dialog);
+
+  /* Use gtk_dialog_run to block execution while the dialog is
+     not closed */
+  while (keep_asking)
+    {
+      char *argv[] = {"/usr/bin/osso-backup", NULL};
+      result = gtk_dialog_run (GTK_DIALOG (dialog));
+      switch (result)
+	{
+	case GTK_RESPONSE_OK:
+	  /* Do nothing, just continue with the process */	  
+	  keep_asking = FALSE;
+	  break;
+	    
+	case GTK_RESPONSE_CANCEL:
+	  keep_asking = FALSE;
+	  break;
+	    
+	case HAM_BACKUP_RESPONSE:
+ 	  run_cmd (argv, ip_suggest_backup_cmd_done, NULL);
+	  break;
+	    
+	default:
+	  g_assert_not_reached ();
+	}
+    }
+
+  pop_dialog (GTK_WIDGET (dialog));
+  gtk_widget_destroy (GTK_WIDGET (dialog));
+
+  return (result == GTK_RESPONSE_OK);
+}
+
+static void
+ip_suggest_backup_cmd_done (int status, void *data)
+{
+  if (status == -1 || !WIFEXITED (status))
+    {
+      what_the_fock_p ();
+      add_log ("Could not launch backup application.\n");
+    }
+}
+
+static void
+ip_close_apps (bool res, void *data)
+{
+  ip_clos *c = (ip_clos *)data;
+
+  if (res)
+    {
+      close_apps ();
+
+      /* Make sure we are done before killing apps, also makes things look
+       * smoother. */
+      while (gtk_events_pending ())
+	gtk_main_iteration ();
+
+      ip_install_cur (c);
+    }
+  else
+    ip_end (c);
+}
+
 static void
 ip_abort_cur_with_status_details (ip_clos *c)
 {
@@ -780,6 +1009,9 @@ ip_abort_cur (ip_clos *c, const char *msg, bool with_details)
 
   GtkWidget *dialog;
 
+  stop_entertaining_user ();
+  c->entertaining = false;
+
   // XXX - get the button texts correct, etc.
 
   if (with_details)
@@ -787,7 +1019,7 @@ ip_abort_cur (ip_clos *c, const char *msg, bool with_details)
       if (is_last)
 	{
 	  dialog = hildon_note_new_confirmation_add_buttons 
-	    (get_dialog_parent (),
+	    (NULL,
 	     msg,
 	     _("ai_ni_bd_details"), 1,
 	     _("ai_ni_bd_close"), GTK_RESPONSE_CANCEL,
@@ -796,11 +1028,11 @@ ip_abort_cur (ip_clos *c, const char *msg, bool with_details)
       else
 	{
 	  dialog = hildon_note_new_confirmation_add_buttons 
-	    (get_dialog_parent (),
+	    (NULL,
 	     msg,
-	     _("Continue"), GTK_RESPONSE_OK,
+	     _("ai_bd_continue"), GTK_RESPONSE_OK,
 	     _("ai_ni_bd_details"), 1,
-	     _("ai_ni_bd_close"), GTK_RESPONSE_CANCEL,
+	     _("ai_bd_notice_cancel"), GTK_RESPONSE_CANCEL,
 	     NULL);
 	}
     }
@@ -809,7 +1041,7 @@ ip_abort_cur (ip_clos *c, const char *msg, bool with_details)
       if (is_last)
 	{
 	  dialog = hildon_note_new_confirmation_add_buttons 
-	    (get_dialog_parent (),
+	    (NULL,
 	     msg,
 	     _("ai_ni_bd_close"), GTK_RESPONSE_CANCEL,
 	     NULL);
@@ -817,15 +1049,15 @@ ip_abort_cur (ip_clos *c, const char *msg, bool with_details)
       else
 	{
 	  dialog = hildon_note_new_confirmation_add_buttons 
-	    (get_dialog_parent (),
+	    (NULL,
 	     msg,
-	     _("Continue"), GTK_RESPONSE_OK,
-	     _("ai_ni_bd_close"), GTK_RESPONSE_CANCEL,
+	     _("ai_bd_continue"), GTK_RESPONSE_OK,
+	     _("ai_bd_notice_cancel"), GTK_RESPONSE_CANCEL,
 	     NULL);
 	}
     }
 
-  push_dialog_parent (dialog);
+  push_dialog (dialog);
 
   g_signal_connect (dialog, "response",
 		    G_CALLBACK (ip_abort_response), c);
@@ -841,11 +1073,20 @@ ip_abort_response (GtkDialog *dialog, gint response, gpointer data)
     ip_show_cur_problem_details (c);
   else
     {
-      pop_dialog_parent (GTK_WIDGET (dialog));
+      pop_dialog (GTK_WIDGET (dialog));
       gtk_widget_destroy (GTK_WIDGET (dialog));
       
       if (response == GTK_RESPONSE_OK)
-	ip_install_next (c);
+	{
+	  /* We only get an OK response when there is another package
+	     to install.  Thus, we start the entertainment here again
+	     since we know that some action will happen.
+	  */
+	  start_entertaining_user ();
+	  c->entertaining = true;
+
+	  ip_install_next (c);
+	}
       else
 	ip_end (c);
     }
@@ -856,17 +1097,40 @@ ip_end (void *data)
 {
   ip_clos *c = (ip_clos *)data;
 
+  if (c->entertaining)
+    stop_entertaining_user ();
+
   get_package_list (APTSTATE_DEFAULT);
   save_backup_data ();
 
   if (c->reboot)
-    annoy_user_with_cont (_("You should reboot now"),
-			  c->cont, c->data);
+    annoy_user (_("ai_ni_device_restart"), 
+		ip_end_rebooting,
+		c);
   else
-    c->cont (c->data);
+    ip_end_after_reboot (c);
+}
+
+static void
+ip_end_after_reboot (void *data)
+{
+  ip_clos *c = (ip_clos *)data;
+
+  c->cont (c->n_successful, c->data);
   delete c;
 }
 
+static void
+ip_end_rebooting (void *data)
+{
+  /* Do normal end */
+  ip_end_after_reboot (data);
+
+  /* Reboot the device */
+  send_reboot_message ();
+}
+
+
 
 /* UNINSTALL_PACKAGE - Overview
 
@@ -998,7 +1262,7 @@ up_checkrm_cmd_done (int status, void *data)
       char *str =
 	g_strdup_printf (_("ai_ni_error_uninstall_applicationrunning"),
 			 c->pi->get_display_name (true));
-      annoy_user_with_cont (str, up_end, c);
+      annoy_user (str, up_end, c);
       g_free (str);
     }
   else
@@ -1013,24 +1277,29 @@ up_remove (up_clos *c)
       add_log ("-----\n");
       add_log ("Uninstalling %s %s\n", c->pi->name, c->pi->installed_version);
       
-      show_progress (_("ai_nw_uninstalling"));
+      char *title = g_strdup_printf (_("ai_nw_uninstalling"),
+				     c->pi->get_display_name (true));
+      set_entertainment_fun (NULL, -1, 0);
+      set_entertainment_cancel (NULL, NULL);
+      set_entertainment_title (title);
+      g_free (title);
+
+      start_entertaining_user ();
+
       apt_worker_remove_package (c->pi->name, up_remove_reply, c);
     }
   else
     {
       if (c->pi->info.removable_status == status_system_update_unremovable)
-	annoy_user_with_cont
-	  (_("ai_ni_error_system"), up_end, c);
-      if (c->pi->info.removable_status == status_needed)
-	annoy_user_with_details_with_cont
-	  (_("ai_ni_error_uninstall_packagesneeded"),
-	   c->pi, remove_details, up_end, c);
+	annoy_user (_("ai_ni_unable_to_uninstall_system_update"), up_end, c);
+      else if (c->pi->info.removable_status == status_needed)
+	annoy_user_with_details (_("ai_ni_error_uninstall_packagesneeded"),
+				 c->pi, remove_details, up_end, c);
       else
 	{
 	  char *str = g_strdup_printf (_("ai_ni_error_uninstallation_failed"),
 				       c->pi->get_display_name (true));
-	  annoy_user_with_details_with_cont (str, c->pi, remove_details,
-					     up_end, c);
+	  annoy_user_with_details (str, c->pi, remove_details, up_end, c);
 	  g_free (str);
 	}
     }
@@ -1041,7 +1310,7 @@ up_remove_reply (int cmd, apt_proto_decoder *dec, void *data)
 {
   up_clos *c = (up_clos *)data;
 
-  hide_progress ();
+  stop_entertaining_user ();
 
   if (dec == NULL)
     {
@@ -1057,14 +1326,14 @@ up_remove_reply (int cmd, apt_proto_decoder *dec, void *data)
     {
       char *str = g_strdup_printf (_("ai_ni_uninstall_successful"),
 				   c->pi->get_display_name (true));
-      annoy_user_with_cont (str, up_end, c);
+      annoy_user (str, up_end, c);
       g_free (str);
     }
   else
     {
       char *str = g_strdup_printf (_("ai_ni_error_uninstallation_failed"),
 				   c->pi->get_display_name (true));
-      annoy_user_with_cont (str, up_end, c);
+      annoy_user (str, up_end, c);
       g_free (str);
     }
 }
@@ -1079,7 +1348,14 @@ up_end (void *data)
 }
 
 
-/* INSTALL_LOCAL_DEB_FILE - Overview
+/* INSTALL_FILE - Overview
+
+   0. Localize file
+
+   1. Dispatch on extenstion to either if_install_local_deb_file or
+      open_local_install_instructions.
+
+   IF_INSTALL_LOCAL_DEB_FILE:
 
    0. Get details of file.
 
@@ -1098,33 +1374,70 @@ up_end (void *data)
  */
 
 struct if_clos {
-  const char *filename;
+  char *filename;
 
   package_info *pi;
 
-  void (*cont) (void *);
+  void (*cont) (bool, void *);
   void *data;
 };
 
+static void if_local (char *local_filename, void *data);
+
+static void if_install_local_deb_file (void *data);
+
 static void if_details_reply (int cmd, apt_proto_decoder *dec, void *data);
 static void if_show_legalese (bool res, void *data);
 static void if_install (bool res, void *data);
 static void if_install_reply (int cmd, apt_proto_decoder *dec, void *data);
 static void if_fail (bool res, void *data);
-static void if_end (void *data);
+static void if_end_with_failure (void *data);
+static void if_end_with_success (void *data);
+static void if_end (bool success, void *data);
 
-void
-install_local_deb_file (const char *filename,
-			void (*cont) (void *data), void *data)
+void install_file (const char *filename,
+		   void (*cont) (bool success, void *data), void *data)
 {
   if_clos *c = new if_clos;
 
-  c->filename = filename;
+  c->filename = NULL;
   c->pi = NULL;
   c->cont = cont;
   c->data = data;
 
-  fprintf (stderr, "INSTALL DEB: %s\n", filename);
+  localize_file_and_keep_it_open (filename, if_local, c);
+}
+
+static void
+if_local (char *local_filename, void *data)
+{
+  if_clos *c = (if_clos *)data;
+
+  if (local_filename)
+    {
+      if (g_str_has_suffix (local_filename, ".install"))
+	{
+	  /* XXX - if_end_with_success is too optimistic but
+	           open_local_install_instructions doesn't report
+	           the results yet.
+	  */
+	  open_local_install_instructions (local_filename,
+					   if_end_with_success, c);
+	}
+      else
+	{
+	  c->filename = local_filename;
+	  if_install_local_deb_file (c);
+	}
+    }
+  else
+    if_end (false, c);
+}
+
+void
+if_install_local_deb_file (void *data)
+{
+  if_clos *c = (if_clos *)data;
 
   apt_worker_get_file_details (!(red_pill_mode && red_pill_show_all),
 			       c->filename, if_details_reply, c);
@@ -1147,7 +1460,7 @@ if_details_reply (int cmd, apt_proto_decoder *dec, void *data)
 
   if (dec == NULL)
     {
-      if_end (c);
+      if_end (false, c);
       return;
     }
 
@@ -1222,7 +1535,7 @@ if_show_legalese (bool res, void *data)
   if (res)
     scare_user_with_legalese (false, if_install, c);
   else
-    if_end (c);
+    if_end (false, c);
 }
 
 static void
@@ -1232,15 +1545,23 @@ if_install (bool res, void *data)
 
   if (res)
     {
-      show_progress (c->pi->installed_version
-		     ? _("ai_nw_updating")
-		     : _("ai_nw_installing"));
+      char *title = g_strdup_printf ((c->pi->installed_version
+				      ? _("ai_nw_updating")
+				      : _("ai_nw_installing")),
+				     c->pi->get_display_name (false));
+      set_entertainment_fun (NULL, -1, 0);
+      set_entertainment_cancel (NULL, NULL);
+      set_entertainment_title (title);
+      g_free (title);
+
+      start_entertaining_user ();
+
       set_log_start ();
       apt_worker_install_file (c->filename,
 			       if_install_reply, c);
     }
   else
-    if_end (c);
+    if_end (false, c);
 }
 
 static void
@@ -1248,11 +1569,11 @@ if_install_reply (int cmd, apt_proto_decoder *dec, void *data)
 {
   if_clos *c = (if_clos *)data;
 
-  hide_progress ();
+  stop_entertaining_user ();
 
   if (dec == NULL)
     {
-      if_end (c);
+      if_end (false, c);
       return;
     }
 
@@ -1267,7 +1588,7 @@ if_install_reply (int cmd, apt_proto_decoder *dec, void *data)
 				   ? _("ai_ni_update_successful")
 				   : _("ai_ni_install_successful"),
 				   c->pi->get_display_name (false));
-      annoy_user_with_cont (str, if_end, c);
+      annoy_user (str, if_end_with_success, c);
       g_free (str);
     }
   else
@@ -1283,7 +1604,7 @@ if_install_reply (int cmd, apt_proto_decoder *dec, void *data)
 			       : _("ai_ni_error_installation_failed"),
 			       c->pi->get_display_name (false));
 
-      annoy_user_with_cont (msg, if_end, c);
+      annoy_user (msg, if_end_with_failure, c);
       g_free (msg);
     }
 }
@@ -1299,24 +1620,44 @@ if_fail (bool res, void *data)
       bool with_details;
       installable_status_to_message (c->pi, msg, with_details);
       if (with_details)
-	annoy_user_with_details_with_cont (msg, c->pi, install_details, 
-					   if_end, c);
+	annoy_user_with_details (msg, c->pi, install_details,
+				 if_end_with_failure, c);
       else
-	annoy_user_with_cont (msg, if_end, c);
+	annoy_user (msg, if_end_with_failure, c);
       g_free (msg);
     }
   else
-    if_end (c);
+    if_end (false, c);
 }
 
 static void
-if_end (void *data)
+if_end_with_success (void *data)
 {
   if_clos *c = (if_clos *)data;
   
+  if_end (true, c);
+}
+
+static void
+if_end_with_failure (void *data)
+{
+  if_clos *c = (if_clos *)data;
+  
+  if_end (false, c);
+}
+
+static void
+if_end (bool success, void *data)
+{
+  if_clos *c = (if_clos *)data;
+  
+  cleanup_temp_file ();
+
+  g_free (c->filename);
+
   if (c->pi)
     c->pi->unref ();
 
-  c->cont (c->data);
+  c->cont (success, c->data);
   delete c;
 }
