@@ -91,7 +91,7 @@ add_entry (GtkWidget *box, GtkSizeGroup *group,
 static void
 pill_response (GtkDialog *dialog, gint response, gpointer unused)
 {
-  pop_dialog_parent (GTK_WIDGET (dialog));
+  pop_dialog (GTK_WIDGET (dialog));
   gtk_widget_destroy (GTK_WIDGET (dialog));
 
   if (red_pill_mode != (response == GTK_RESPONSE_YES))
@@ -108,12 +108,12 @@ ask_the_pill_question ()
   GtkWidget *dialog;
 
   dialog =
-    hildon_note_new_confirmation_add_buttons (get_dialog_parent (), 
+    hildon_note_new_confirmation_add_buttons (NULL, 
 					      "Which pill?",
 					      "Red", GTK_RESPONSE_YES,
 					      "Blue", GTK_RESPONSE_NO,
 					      NULL);
-  push_dialog_parent (dialog);
+  push_dialog (dialog);
   g_signal_connect (dialog, "response",
 		    G_CALLBACK (pill_response), NULL);
   gtk_widget_show_all (dialog);
@@ -140,7 +140,7 @@ get_catalogues_callback (int cmd,
     }
   else
     {
-      annoy_user (_("ai_ni_operation_failed"));
+      what_the_fock_p ();
       xexp_free (x);
       c->cont (NULL, c->data);
     }
@@ -157,84 +157,6 @@ get_catalogues (void (*cont) (xexp *catalogues, void *data),
   apt_worker_get_catalogues (get_catalogues_callback, c);
 }
 
-struct set_catalogues_closure {
-  int state;
-  bool refresh;
-  bool ask;
-  void (*cont) (bool res, void *data);
-  void *data;
-};
-
-static void
-set_catalogues_reply (int cmd, apt_proto_decoder *dec, void *data)
-{
-  set_catalogues_closure *c = (set_catalogues_closure *)data;
-  int success = 0;
-
-  if (dec)
-    success = dec->decode_int ();
-
-  if (!success)
-    {
-      if (dec)
-	annoy_user_with_log (_("ai_ni_operation_failed"));
-      c->cont (false, c->data);
-      delete c;
-      return;
-    }
-
-  if (c->refresh)
-    {
-      refresh_package_cache_with_cont (c->state, c->ask,
-				       c->cont, c->data);
-      delete c;
-    }
-  else
-    {
-      c->cont (true, c->data);
-      delete c;
-    }
-
-  save_backup_data ();
-}
-
-static void
-set_catalogues_1 (int state,
-		  xexp *catalogues, bool refresh, bool ask,
-		  void (*cont) (bool res, void *data),
-		  void *data)
-{
-  set_catalogues_closure *c = new set_catalogues_closure;
-  c->state = state;
-  c->ask = ask;
-  c->refresh = refresh;
-  c->cont = cont;
-  c->data = data;
-
-  apt_worker_set_catalogues (state, catalogues,
-			     set_catalogues_reply, c);
-}
-
-void
-set_catalogues (xexp *catalogues, bool refresh, bool ask,
-		void (*cont) (bool res, void *data),
-		void *data)
-{
-  set_catalogues_1 (APTSTATE_DEFAULT, catalogues,
-		    refresh, ask,
-		    cont, data);
-}
-
-void
-set_temp_catalogues (xexp *catalogues,
-		     void (*cont) (bool res, void *data),
-		     void *data)
-{
-  set_catalogues_1 (APTSTATE_TEMP, catalogues,
-		    true, false,
-		    cont, data);
-}
-
 const char *
 catalogue_name (xexp *x)
 {
@@ -275,17 +197,25 @@ struct catcache {
   xexp *catalogue_xexp;
   bool enabled, readonly, foreign;
   const char *name;
+  char *detail;
 };
 
 struct cat_dialog_closure {
   catcache *caches;
   xexp *catalogues_xexp;
   bool dirty;
+  bool show_errors;
+
+  catcache *selected_cat;
+  GtkTreeIter selected_iter;
 
   GtkTreeView *tree;
   GtkListStore *store;
   GtkWidget *edit_button;
   GtkWidget *delete_button;
+
+  void (*cont) (bool changed, void *data);
+  void *data;
 };
 
 struct cat_edit_closure {
@@ -359,7 +289,7 @@ cat_edit_response (GtkDialog *dialog, gint response, gpointer clos)
 
   delete c;
  
-  pop_dialog_parent (GTK_WIDGET (dialog));
+  pop_dialog (GTK_WIDGET (dialog));
   gtk_widget_destroy (GTK_WIDGET (dialog));
 
   if (should_ask_the_pill_question)
@@ -398,7 +328,7 @@ show_cat_edit_dialog (cat_dialog_closure *cat_dialog, xexp *catalogue,
     {
       GtkWidget *button;
 
-      dialog = gtk_dialog_new_with_buttons (title, get_dialog_parent (),
+      dialog = gtk_dialog_new_with_buttons (title, NULL,
 					    GTK_DIALOG_MODAL,
 					    NULL);
 
@@ -410,7 +340,7 @@ show_cat_edit_dialog (cat_dialog_closure *cat_dialog, xexp *catalogue,
       gtk_widget_grab_focus (button);
     }
   else
-    dialog = gtk_dialog_new_with_buttons (title, get_dialog_parent (),
+    dialog = gtk_dialog_new_with_buttons (title, NULL,
 					  GTK_DIALOG_MODAL,
 					  _("ai_bd_new_repository_ok"),
 					  GTK_RESPONSE_OK,
@@ -418,7 +348,7 @@ show_cat_edit_dialog (cat_dialog_closure *cat_dialog, xexp *catalogue,
 					  GTK_RESPONSE_CANCEL,
 					  NULL);
 
-  push_dialog_parent (dialog);
+  push_dialog (dialog);
 
   gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
 
@@ -519,9 +449,24 @@ cat_text_func (GtkTreeViewColumn *column,
 	       GtkTreeIter *iter,
 	       gpointer data)
 {
+  cat_dialog_closure *cd = (cat_dialog_closure *)data;
   catcache *c;
+
   gtk_tree_model_get (model, iter, 0, &c, -1);
-  g_object_set (cell, "text", c? c->name : NULL, NULL);
+  
+  if (cd->show_errors
+      && c != NULL
+      && c->detail
+      && cd->selected_cat == c)
+    {
+      gchar *markup;
+      markup = g_markup_printf_escaped ("%s\n<small>%s</small>",
+					c->name, c->detail);
+      g_object_set (cell, "markup", markup, NULL);
+      g_free (markup);
+    }
+  else
+    g_object_set (cell, "text", c? c->name : NULL, NULL);
 }
 
 static void
@@ -540,11 +485,24 @@ cat_row_activated (GtkTreeView *treeview,
       if (c == NULL)
 	return;
 
+      if (c->readonly)
+	return;
+
       show_cat_edit_dialog (c->cat_dialog, c->catalogue_xexp,
-			    false, c->readonly);
+			    false, false);
     }
 }
 
+static void
+emit_row_changed (GtkTreeModel *model, GtkTreeIter *iter)
+{
+  GtkTreePath *path;
+
+  path = gtk_tree_model_get_path (model, iter);
+  g_signal_emit_by_name (model, "row-changed", path, iter);
+  gtk_tree_path_free (path);
+}
+
 static void
 cat_selection_changed (GtkTreeSelection *selection, gpointer data)
 {
@@ -552,23 +510,60 @@ cat_selection_changed (GtkTreeSelection *selection, gpointer data)
   GtkTreeIter iter;
   cat_dialog_closure *c = (cat_dialog_closure *)data;
 
-  if (gtk_tree_selection_get_selected (selection, &model, &iter))
-    {
-      catcache *cat;
-      gtk_tree_model_get (model, &iter, 0, &cat, -1);
-      if (cat == NULL)
-	return;
+  catcache *old_selected = c->selected_cat;
+  catcache *new_selected;
 
-      gtk_widget_set_sensitive (c->edit_button, !cat->foreign);
-      gtk_widget_set_sensitive (c->delete_button, !cat->readonly);
+  if (gtk_tree_selection_get_selected (selection, &model, &iter))
+    gtk_tree_model_get (model, &iter, 0, &new_selected, -1);
+  else
+    new_selected = NULL;
+
+  c->selected_cat = new_selected;
+  if (old_selected)
+    emit_row_changed (model, &c->selected_iter);
+  c->selected_iter = iter;
+
+  if (new_selected)
+    {
+      emit_row_changed (model, &iter);
+      gtk_widget_set_sensitive (c->edit_button,
+				(!new_selected->readonly
+				 && !new_selected->foreign));
+      if (!c->show_errors)
+	gtk_widget_set_sensitive (c->delete_button,
+				  !new_selected->readonly);
     }
   else
     {
       gtk_widget_set_sensitive (c->edit_button, FALSE);
-      gtk_widget_set_sensitive (c->delete_button, FALSE);
+      if (!c->show_errors)
+	gtk_widget_set_sensitive (c->delete_button, FALSE);
     }
 }
 
+static char *
+render_catalogue_errors (xexp *cat)
+{
+  xexp *errors = xexp_aref (cat, "errors");
+  if (errors == NULL || xexp_first (errors) == NULL)
+    return NULL;
+
+  GString *report = g_string_new ("");
+
+  for (xexp *err = xexp_first (errors); err; err = xexp_rest (err))
+    {
+      g_string_append_printf (report, "%s\n %s",
+			      xexp_aref_text (err, "uri"),
+			      xexp_aref_text (err, "msg"));
+      if (xexp_rest (err))
+	g_string_append (report, "\n");
+    }
+
+  char *str = report->str;
+  g_string_free (report, 0);
+  return str;
+}
+
 static catcache *
 make_catcache_from_xexp (cat_dialog_closure *c, xexp *x)
 {
@@ -581,6 +576,7 @@ make_catcache_from_xexp (cat_dialog_closure *c, xexp *x)
       cat->readonly = xexp_aref_bool (x, "essential");
       cat->foreign = false;
       cat->name = catalogue_name (x);
+      cat->detail = render_catalogue_errors (x);
     }
   else if (xexp_is (x, "source") && xexp_is_text (x))
     {
@@ -588,6 +584,7 @@ make_catcache_from_xexp (cat_dialog_closure *c, xexp *x)
       cat->readonly = true;
       cat->foreign = true;
       cat->name = xexp_text (x);
+      cat->detail = NULL;
     }
   else
     {
@@ -605,9 +602,11 @@ reset_cat_list (cat_dialog_closure *c)
   for (catcache *cat = c->caches; cat; cat = next)
     {
       next = cat->next;
+      delete cat->detail;
       delete cat;
     }
   c->caches = NULL;
+  c->selected_cat = NULL;
 }
 
 static void
@@ -618,6 +617,15 @@ set_cat_list (cat_dialog_closure *c)
   for (xexp *catx = xexp_first (c->catalogues_xexp); catx;
        catx = xexp_rest (catx))
     {
+      xexp *errors = (xexp_is (catx, "catalogue")
+		      ? xexp_aref (catx, "errors")
+		      : NULL);
+
+      if (c->show_errors
+	  && (errors == NULL
+	      || xexp_first (errors) == NULL))
+	continue;
+
       catcache *cat = make_catcache_from_xexp (c, catx);
       if (cat)
 	{
@@ -628,6 +636,11 @@ set_cat_list (cat_dialog_closure *c)
 					     position,
 					     0, cat,
 					     -1);
+	  if (position == 0)
+	    {
+	      c->selected_cat = cat;
+	      c->selected_iter = iter;
+	    }
 	  position += 1;
 	}
     }
@@ -652,21 +665,23 @@ make_cat_list (cat_dialog_closure *c)
     GTK_TREE_VIEW (gtk_tree_view_new_with_model (GTK_TREE_MODEL (c->store)));
 
   renderer = gtk_cell_renderer_pixbuf_new ();
+  g_object_set (renderer, "yalign", 0.0, NULL);
   gtk_tree_view_insert_column_with_data_func (c->tree,
 					      -1,
 					      NULL,
 					      renderer,
 					      cat_icon_func,
-					      c->tree,
+					      c,
 					      NULL);
 
   renderer = gtk_cell_renderer_text_new ();
+  g_object_set (renderer, "yalign", 0.0, NULL);
   gtk_tree_view_insert_column_with_data_func (c->tree,
 					      -1,
 					      NULL,
 					      renderer,
 					      cat_text_func,
-					      c->tree,
+					      c,
 					      NULL);
 
   g_signal_connect (c->tree, "row-activated", 
@@ -710,11 +725,6 @@ remove_cat_cont (bool res, void *data)
   delete c;
 }
 
-static void
-ignore_result (bool res, void *data)
-{
-}
-
 #define REPO_RESPONSE_NEW    1
 #define REPO_RESPONSE_EDIT   2
 #define REPO_RESPONSE_REMOVE 3
@@ -727,7 +737,9 @@ cat_response (GtkDialog *dialog, gint response, gpointer clos)
   if (response == REPO_RESPONSE_NEW)
     {
       xexp *x = xexp_list_new ("catalogue");
-      xexp_cons (x, xexp_text_new ("name", ""));
+      xexp_aset_text (x, "name", "");
+      xexp_aset_text (x, "uri", "http://");
+      xexp_aset_text (x, "components", "user");
       show_cat_edit_dialog (c, x, true, false);
       return;
     }
@@ -761,16 +773,12 @@ cat_response (GtkDialog *dialog, gint response, gpointer clos)
 
   if (response == GTK_RESPONSE_CLOSE)
     {
-      if (c->dirty)
-	set_catalogues (c->catalogues_xexp, true, true,
-			ignore_result, NULL);
-
       reset_cat_list (c);
-      xexp_free (c->catalogues_xexp);
-
-      delete c;
-      pop_dialog_parent (GTK_WIDGET (dialog));
+      pop_dialog (GTK_WIDGET (dialog));
       gtk_widget_destroy (GTK_WIDGET (dialog));
+
+      c->cont (c->dirty, c->data);
+      delete c;
     }
 }
 
@@ -788,60 +796,129 @@ insensitive_cat_delete_press (GtkButton *button, gpointer data)
 }
 
 void
-show_cat_dialog_with_catalogues (xexp *catalogues, void *unused)
+show_catalogue_dialog (xexp *catalogues,
+		       bool show_errors,
+		       void (*cont) (bool changed, void *data),
+		       void *data)
 {
-  if (catalogues == NULL)
-    return;
+  g_return_if_fail (catalogues != NULL);
 
   cat_dialog_closure *c = new cat_dialog_closure;
   c->caches = NULL;
   c->catalogues_xexp = catalogues;
   c->dirty = false;
+  c->show_errors = show_errors;
+  c->selected_cat = NULL;
+  c->cont = cont;
+  c->data = data;
+
+  xexp_write (stdout, catalogues);
 
   GtkWidget *dialog = gtk_dialog_new ();
 
-  gtk_window_set_title (GTK_WINDOW (dialog), _("ai_ti_repository"));
-  gtk_window_set_transient_for (GTK_WINDOW (dialog), get_dialog_parent ());
-  gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
-  push_dialog_parent (dialog);
+  if (show_errors)
+    gtk_window_set_title (GTK_WINDOW (dialog), _("ai_ti_failed_repositories"));
+  else
+    gtk_window_set_title (GTK_WINDOW (dialog), _("ai_ti_repository"));
+  
+  push_dialog (dialog);
   
   gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
-  gtk_dialog_add_button (GTK_DIALOG (dialog), 
-			 _("ai_bd_repository_new"), REPO_RESPONSE_NEW);
+
+  if (!show_errors)
+    gtk_dialog_add_button (GTK_DIALOG (dialog), 
+			   _("ai_bd_repository_new"), REPO_RESPONSE_NEW);
+
   c->edit_button = 
     gtk_dialog_add_button (GTK_DIALOG (dialog), 
 			   _("ai_bd_repository_edit"), REPO_RESPONSE_EDIT);
-  c->delete_button =
-    gtk_dialog_add_button (GTK_DIALOG (dialog), 
-			   _("ai_bd_repository_delete"), REPO_RESPONSE_REMOVE);
+
+  if (!show_errors)
+    c->delete_button =
+      gtk_dialog_add_button (GTK_DIALOG (dialog), 
+			     _("ai_bd_repository_delete"), REPO_RESPONSE_REMOVE);
+
   gtk_dialog_add_button (GTK_DIALOG (dialog),
 			 _("ai_bd_repository_close"), GTK_RESPONSE_CLOSE);
   respond_on_escape (GTK_DIALOG (dialog), GTK_RESPONSE_CLOSE);
   
-  g_signal_connect (c->delete_button, "insensitive_press",
-		    G_CALLBACK (insensitive_cat_delete_press), c);
-  
-  set_dialog_help (dialog, AI_TOPIC ("repository"));
-  
   gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
 			       make_cat_list (c));
-  
-  gtk_widget_set_sensitive (c->edit_button, FALSE);
-  gtk_widget_set_sensitive (c->delete_button, FALSE);
-  
-  gtk_widget_set_usize (dialog, 0, 250);
-  
+
+  gtk_widget_set_sensitive (c->edit_button, FALSE); 
+  if (!show_errors)
+    {
+      gtk_widget_set_sensitive (c->delete_button, FALSE);
+      g_signal_connect (c->delete_button, "insensitive_press",
+			G_CALLBACK (insensitive_cat_delete_press), c);
+
+      set_dialog_help (dialog, AI_TOPIC ("repository"));
+    }
+    
   g_signal_connect (dialog, "response",
 		    G_CALLBACK (cat_response), c);
+
+  gtk_widget_set_usize (dialog, 600, 300);
+  
   gtk_widget_show_all (dialog);
 }
 
+/* "Application Catalogues" interaction flow.
+ */
+
+struct scdf_clos {
+  xexp *catalogues;
+};
+
+static void scdf_with_catalogues (xexp *catalogues, void *data);
+static void scdf_dialog_done (bool changed, void *date);
+static void scdf_end (bool res, void *data);
+
 void
-show_repo_dialog ()
+show_catalogue_dialog_flow ()
 {
-  get_catalogues (show_cat_dialog_with_catalogues, NULL);
+  if (start_interaction_flow ())
+    {
+      scdf_clos *c = new scdf_clos;
+      get_catalogues (scdf_with_catalogues, c);
+    }
 }
 
+static void
+scdf_with_catalogues (xexp *catalogues, void *data)
+{
+  scdf_clos *c = (scdf_clos *)data;
+
+  c->catalogues = catalogues;
+  show_catalogue_dialog (catalogues, false,
+			 scdf_dialog_done, c);
+}
+
+static void
+scdf_dialog_done (bool changed, void *data)
+{
+  scdf_clos *c = (scdf_clos *)data;
+
+  if (changed)
+    refresh_package_cache (APTSTATE_DEFAULT,
+			   c->catalogues, true, false,
+			   scdf_end, c);
+  else
+    {
+      xexp_free (c->catalogues);
+      scdf_end (true, c);
+    }
+      
+  c->catalogues = NULL;
+}
+
+static void
+scdf_end (bool res, void *data)
+{
+  end_interaction_flow ();
+}
+
+
 /* Adding catalogues 
  */
 
@@ -936,12 +1013,7 @@ static void
 add_catalogues_cont_4 (bool res, void *data)
 {
   add_catalogues_closure *c = (add_catalogues_closure *)data;
-
-  /* We ignore errors from refreshing the cache here since
-     installation of packages might continue.
-  */
-  c->cont (true, c->data);
-  xexp_free (c->catalogues);
+  c->cont (res, c->data);
   delete c;
 }
 
@@ -958,8 +1030,13 @@ add_catalogues_cont_2 (add_catalogues_closure *c)
 	 the installation script later.
       */
       if (c->catalogues_changed || c->update)
-	set_catalogues (c->catalogues, true, !c->update,
-			add_catalogues_cont_4, c);
+	{
+	  refresh_package_cache (APTSTATE_DEFAULT,
+				 c->catalogues,
+				 !c->update, c->update,
+				 add_catalogues_cont_4, c);
+	  c->catalogues = NULL;
+	}
       else
 	{
 	  c->cont (true, c->data);
@@ -997,7 +1074,11 @@ add_catalogues_cont_2 (add_catalogues_closure *c)
 	      char *str;
 	      const char *name = catalogue_name (c->rest);
 
-	      if (c->update)
+	      if (cont == add_catalogues_cont_3_enable)
+		str = g_strdup_printf ("%s\n%s",
+				       _("ai_ia_add_catalogue_enable"),
+				       name);
+	      else if (c->update)
 		str = g_strdup_printf (_("ai_ia_add_catalogue_text"),
 				       name);
 	      else
@@ -1060,3 +1141,47 @@ add_catalogues (xexp *catalogues,
 
   get_catalogues (add_catalogues_cont_1, c);
 }
+
+GString *
+render_catalogue_report (xexp *catalogue_report)
+{
+  if (catalogue_report == NULL)
+    return g_string_new (_("ai_ni_operation_failed"));
+
+  GString *report = g_string_new ("");
+
+  for (xexp *cat = xexp_first (catalogue_report); cat; cat = xexp_rest (cat))
+    {
+      xexp *errors = xexp_aref (cat, "errors");
+      if (errors == NULL || xexp_first (errors) == NULL)
+	continue;
+
+      g_string_append_printf (report, "%s:\n", catalogue_name (cat));
+      for (xexp *err = xexp_first (errors); err; err = xexp_rest (err))
+	g_string_append_printf (report, "  %s\n   %s\n",
+				xexp_aref_text (err, "uri"),
+				xexp_aref_text (err, "msg"));
+    }
+
+  return report;
+}
+
+xexp *
+get_failed_catalogues (xexp *catalogue_report)
+{
+  xexp *failed_cat_list = xexp_list_new ("catalogues");
+
+  if (catalogue_report == NULL)
+    return failed_cat_list;
+
+  for (xexp *cat = xexp_first (catalogue_report); cat; cat = xexp_rest (cat))
+    {
+      xexp *errors = xexp_aref (cat, "errors");
+      if (errors == NULL || xexp_first (errors) == NULL)
+	continue;
+
+      xexp_append_1 (failed_cat_list, xexp_copy (cat));
+    }
+
+  return failed_cat_list;
+}
