@@ -28,7 +28,12 @@ extern "C" {
 #include "xexp.h"
 }
 
-void show_repo_dialog ();
+void show_catalogue_dialog_flow ();
+
+void show_catalogue_dialog (xexp *catalogues,
+			    bool show_errors,
+			    void (*cont) (bool changed, void *data),
+			    void *data);
 
 /* A 'catalogue' is an association xexp with the following elements:
 
@@ -67,14 +72,6 @@ void show_repo_dialog ();
 void get_catalogues (void (*cont) (xexp *catalogues, void *data),
 		     void *data);
 
-void set_catalogues (xexp *catalogues, bool refresh, bool ask,
-		     void (*cont) (bool res, void *data),
-		     void *data);
-
-void set_temp_catalogues (xexp *catalogues,
-			  void (*cont) (bool res, void *data),
-			  void *data);
-
 void add_catalogues (xexp *catalogues, bool ask, bool update,
 		     void (*cont) (bool res, void *data),
 		     void *data);
