@@ -25,6 +25,7 @@
 #define UTIL_H
 
 #include <gtk/gtk.h>
+#include <gdk/gdkx.h>
 #include <libgnomevfs/gnome-vfs.h>
 
 #include "main.h"
@@ -32,34 +33,61 @@
 /** Interaction flows
 
   The Application Manager can run one interaction flow at any single
-  time.  A flow is started with START_INTERACTION_FLOW and ended with
-  END_INTERACTION_FLOW.
+  time.  A local flow is started with START_INTERACTION_FLOW and ended
+  with END_INTERACTION_FLOW.  A flowe that should run on top of a
+  foreign window is started with START_FOREIGN_INTERACTION_FLOW and
+  ended with END_INTERACTION_FLOW.
 
   When a new interaction flow is to be started while there is already
-  one active, an error message is displayed and START_INTERACTION_FLOW
-  returns false.  The interaction flow should be aborted in that case.
+  one active, START_INTERACTION_FLOW displays an appropriate error
+  message and returns false.  START_FOREIGN_INTERACTION_FLOW does not
+  display this message.
 
-  PUSH_DIALOG_PARENT and POP_DIALOG_PARENT are used to keep track of
-  the top-most window to be used as the parent for new dialogs.
-  GET_DIALOG_PARENT returns that window.  You can pass a NULL window
-  to PUSH_DIALOG_PARENT.  In that case, the flow is system modal,
-  which happens when the flow is started from another application.
-
-  START_INTERACTION_FLOW takes care of the initial PUSH_DIALOG_PARENT
-  and END_INTERACTION_FLOW does the final POP_DIALOG_PARENT.
+  PUSH_DIALOG and POP_DIALOG are used to maintain the stack of dialogs
+  that are currently active.  These calls make sure that the dialogs
+  are properly transient for each other.  When creating a dialog, you
+  should use NULL as the parent.
 
   You must take care to correctly balance calls to
-  START_INTERACTION_FLOW and END_INTERACTION_FLOW, as well as calls to
-  PUSH_DIALOG_PARENT and POP_DIALOG_PARENT.
+  START_INTERACTION_FLOW (or START_FOREIGN_INTERACTION_FLOW) and
+  END_INTERACTION_FLOW, as well as calls to PUSH_DIALOG and
+  POP_DIALOG.
+
+  See "Exiting" for how interaction flows influence the life
+  time of the application.
 */
 
-bool start_interaction_flow (GtkWidget *window);
-void end_interaction_flow (GtkWidget *window);
+bool start_interaction_flow ();
+bool start_foreign_interaction_flow (Window parent);
+void end_interaction_flow ();
 
-void push_dialog_parent (GtkWidget *window);
-void pop_dialog_parent (GtkWidget *window);
-GtkWindow *get_dialog_parent ();
+void push_dialog (GtkWidget *dialog);
+void pop_dialog (GtkWidget *dialog);
 
+/** The main window
+
+  PRESENT_MAIN_WINDOW and HIDE_MAIN_WINDOW will show and hide the main
+  window of the application, respectively.  When it is shown, the main
+  window is also pushed to the front of the window stack.
+
+  See "Exiting" for how the main window influence the life time of the
+  application.
+*/
+
+void present_main_window ();
+void hide_main_window ();
+
+/** Exiting
+
+  The Application Manager can run interaction flows on top of foreign
+  windows and can also show its own window at the same time.  Thus the
+  program should exit when the main window is not shown and no
+  interaction flow is active.  This doesn't happen automatically, but
+  only when MAYBE_EXIT is called.  This way, necessary cleanup actions
+  such as flushing the D-Bus message queue can be performed.
+ */
+
+void maybe_exit ();
 
 /** General dialog helpers
 
@@ -91,23 +119,21 @@ GtkWindow *get_dialog_parent ();
   ANNOY_USER_WITH_DETAILS is like annoy_user but adds a "Details"
   button like ask_yes_no_with_details.
 
-  ANNOY_USER_WITH_LOG is the same as annoy_user.  It used to add a
-  "Log" button that would open the log dialog.  This "Log" button is
-  no longer used in order not to push the scary log into the face of
-  the user too much.  The code, however, still uses
-  annoy_user_with_log in preference to annoy_user when showing an
-  error message where the log is expected to contain more details.
-
   ANNOY_USER_WITH_ERRNO shows a notification that is appropriate for
   the given errno value ERR.  The DETAIL string will be put into the
   Log together with a detailed error message.
 
-  If a 'annoy_user' dialog is already active when any of the
-  annoy_user function is called, no new dialog is displayed.
+  ANNOY_USER_WITH_GNOME_VFS_RESULT does the same for a GnomeVFSResult
+  code instead of a errno code.
 
   IRRITATE_USER shows TEXT in a information banner which goes away
   automatically after a certain time.
 
+  WHAT_THE_FOCK_P irritates the user with a general "Operation failed"
+  message.  Details should appear in the log.  It should be used when
+  a situation has occured that can not reasonably explained to the
+  user, such as the apt-worker returning out-of-sequence replies.
+
   SCARE_USER_WITH_LEGALESE shows one of two legal disclaimers,
   depending on the SURE parameter.  When SURE is true, the disclaimer
   reflects the fact that you know for sure that the user is about to
@@ -136,81 +162,85 @@ void ask_yes_no_with_arbitrary_details (const gchar *title,
 					void (*details) (void *data),
 					void *data);
 
-void annoy_user (const gchar *text);
-void annoy_user_with_cont (const gchar *text, void (*cont) (void *data), void *data);
+void annoy_user (const gchar *text, void (*cont) (void *data), void *data);
 void annoy_user_with_details (const gchar *text,
-			      package_info *pi, detail_kind kind);
-void annoy_user_with_details_with_cont (const gchar *text,
-					package_info *pi, detail_kind kind,
-					void (*cont) (void *data), void *data);
-void annoy_user_with_log (const gchar *text);
+			      package_info *pi, detail_kind kind,
+			      void (*cont) (void *data), void *data);
+void annoy_user_with_arbitrary_details (const gchar *text,
+					void (*details) (void *data),
+					void (*cont) (void *data),
+					void *data);
+void annoy_user_with_arbitrary_details_2 (const gchar *text,
+					  void (*details) (void *data),
+					  void (*cont) (void *data),
+					  void *data);
+
 void annoy_user_with_errno (int err, const gchar *detail,
 			    void (*cont) (void *), void *data);
 void annoy_user_with_gnome_vfs_result (GnomeVFSResult result,
-				       const gchar *detail);
+				       const gchar *detail,
+				       void (*cont) (void *), void *data);
 
 void irritate_user (const gchar *text);
+void what_the_fock_p ();
 
 void scare_user_with_legalese (bool sure,
 			       void (*cont) (bool res, void *data),
 			       void *data);
 
-/** Progress indicator
+/* Progress dialog.
+   
+   START_ENTERTAINING_USER shows a progress bar dialog with a cancel
+   button and STOP_ENTERTAINING_USER removes it.  These two functions
+   must be called in pairs and must be properly nested with respect to
+   push_dialog_parent and pop_dialog_parent.  There will only be at
+   most one progress bar dialog active at any given time, and nesting
+   start_entertaining_user/stop_entertaining_user will just maintain a
+   counter of the nesting depth.
 
-  There is a single global progress indication dialog that can be
-  shown or hidden.  The dialog is application modal and has a "Cancel"
-  button.  The progress bar can either pulse or show a 'completed'
-  fraction.
+   START_ENTERTAINING_USER_SILENTLY is like start_entertaining_user,
+   but does not actually show the progress dialog.  The effect of
+   using start_entertaining_user_silently is to keep a progress dialog
+   open until the corresponding stop_entertaining_user has been
+   called, but to wait for a nested call to start_entertaining_user
+   for actually showing the progress.
 
-  The dialog will be shown when calling SHOW_PROGRESS but also
-  automatically when a status report from the apt-worker comes, except
-  when that status report is for op_updating_cache.  To be prepared
-  for this automatic showing, you should use
-  set_general_progress_title and reset_progress_was_cancelled at the
-  right time.
+   SET_ENTERTAINMENT_FUN sets the state of the progress bar dialog.
+   You can and should call this function before calling
+   start_entertaining_user to setup the initial state of the dialog.
+   TITLE is the text displayed in the dialog and ALREADY and TOTAL
+   determine the content of the progress bar.  When ALREADY is
+   negative, the progress bar pulses.
 
-  XXX - Yes, this logic is ugly and needs to be cleaned up.
+   SET_ENTERTAINMENT_DOWNLOAD_FUN is a specialization of
+   set_entertainment_fun: it automatically provides an appropriate
+   sub-title that includes the total download size.
 
-  SHOW_PROGRESS shows the progress dialog, resets it to pulsing mode,
-  and sets the current operation associated with the dialog to
-  op_general (see apt-worker-proto.h).  The TITLE given in the call to
-  show_progress is used whenever the current operation is op_general.
+   SET_ENTERTAINMENT_CANCEL associates a callback with the "Cancel"
+   button in the dialog.  When CALLBACK is NULL, the button is
+   insensitive.
 
-  SET_GENERAL_PROGRESS_TITLE can be use to set the title used for
-  op_general without causing the dialog to show.
+   CANCEL_ENTERTAINMENT calls the callback associated with
+   set_entertainment_cancel, if there is one.
 
-  SET_PROGRESS sets the displayed progress to relfect that ALREADY
-  units out of TOTAL units have been completed.  When ALREADY is -1,
-  however, the prgress bar starts pulsing.  The current operation is
-  set to OP.  When it is op_general, the last title given to
-  show_progress is used; otherwise, a title appropriate for OP is
-  used.  When OP is op_downloading, the unit for TOTAL is taken to be
-  bytes and this size is included in the title.
+   ENTERTAINMENT_WAS_CANCELLED returns true when cancel_entertainment
+   has been called since the last call to start_entertaining_user.
+   Clicking on the "Cancel" button will also call
+   cancel_entertainment.
+ */
 
-  When the current operation is op_downloading, clicking on the cancel
-  button will call cancel_apt_worker (see apt-worker-client.h).
-  Otherwise, only an information banner saying "Unableto cancel now.
-  Please wait." is shown.
+void start_entertaining_user ();
+void start_entertaining_user_silently ();
+void stop_entertaining_user ();
 
-  HIDE_PROGRESS will hide the progress dialog.  Since the dialog is
-  application modal, it is important to eventually hide it.
-  Otherwise, the application will effectively freeze.
+void set_entertainment_title (const char *main_title);
+void set_entertainment_fun (const char *sub_title,
+			    int64_t alreday, int64_t total);
+void set_entertainment_download_fun (int64_t already, int64_t total);
 
-  PROGRESS_WAS_CANCELLED can be used to determine whether
-  cancel_apt_worker was indeed called as the response to a click on
-  the "Cancel" button (since the last call to show_progress or
-  reset_progress_was_cancelled).
-
-  RESET_PROGRESS_WAS_CANCELLED resets the flag to false that tracks
-  whether the user has clicked on the "Cancel" button.
-*/
-
-void show_progress (const char *title);
-void set_general_progress_title (const char *title);
-void set_progress (apt_proto_operation op, int already, int total);
-bool progress_was_cancelled ();
-void reset_progress_was_cancelled ();
-void hide_progress ();
+void set_entertainment_cancel (void (*callback)(void *), void *data);
+void cancel_entertainment ();
+bool entertainment_was_cancelled ();
 
 /* SHOW_UPDATING and HIDE_UPDATING determine whether the "Updating"
    animation banner should be shown.  They maintain a counter;
@@ -293,7 +323,7 @@ GtkWidget *make_global_package_list (GList *packages,
 				     package_info_callback *activated);
 void clear_global_package_list ();
 void global_package_info_changed (package_info *pi);
-
+void reset_global_target_path ();
 
 /* Global section list widget
 
@@ -330,6 +360,7 @@ void clear_global_section_list ();
  */
 
 void select_package_list (GList *package_list,
+			  int state,
 			  const gchar *title,
 			  const gchar *question,
 			  void (*cont) (gboolean res, GList * pl, void *data),
@@ -413,6 +444,10 @@ void run_cmd (char **argv,
 	      void (*cont) (int status, void *data),
 	      void *data);
 
+/* CLOSE_APPS kills all the user applications currently running in the
+   device except the hildon-application-manager */
+void close_apps (void);
+
 /* Skip over the leading whitespace characters of STR and return a
    pointer to the first non-whitespace one.
 */
@@ -466,11 +501,6 @@ void *pop (GSList *&ptr);
 */
 const char *gettext_alt (const char *id, const char *english);
 
-/* Return the device name.
- */
-const char *device_name ();
-void setup_dbus();
-
 /* Set up a handler that emits the given RESPONSE for DIALOG when
    the user hits Escape.
 */
@@ -485,4 +515,7 @@ void grab_focus_on_map (GtkWidget *widget);
  */
 int64_t get_free_space ();
 
+/* Set a DBUS message to reboot the device */
+void send_reboot_message (void);
+
 #endif /* !UTIL_H */
